/*--------------------------------------------------------------------
 * $Id: atmosphere.c,v 1.68 2006/01/09 14:09:55 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>
#include <ckdfu.h>
#include "uvspec.h"
#include "cloud.h"
#include "../src_c/ascii.h"
#include "../src_c/radiative.h"
#include "../src_f/f77-uscore.h"

#define EPSILON 1E-6

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif


#define M_H2O 18.015
#define N_MOL_CKDFU 6.0235e23



/************************************/
/* prototypes of internal functions */
/************************************/

static int read_atmosphere (char *filename, atm_out_struct *out, 
                            int interpol_method_press, int interpol_method_temper, int *interp_method_gas,
                            int cdf, int quiet, int verbose);

/* static int read_density (char *filename, */
/* 			 atm_out_struct *atm_out, int dens_id, int *interpol_method, */
/* 			 int quiet, int verbose);  */

static int read_and_combine_density (char *filename, atm_out_struct *out,
				     int dens_id, int *interpol_method_gas,
				     float *dens_air_all, float *z_all, int nlev_all,
                                     float *dens_air_atm, float *z_atm, int nlev_atm,
				     int solver, int quiet, int verbose);

static int read_and_convert_rh      (char *filename,
				     float **rh, float *z_all, int nlev_all,
				     int interpol_method_rh, 
                                     float *t, float *h2o, int quiet);

static int calculate_or_read_refind (char *filename,
				     float **refind, float *z_all, int nlev_all,
				     int interpol_method_refind, 
                                     float *t, float *p,
                                     int quiet);

static int combine_profiles (float **dens, float *z, int nlev, int sza,
                             float *dens_all, float *z_all, int nlev_all, 
                             int interpol_method_dens,
                             float ***dens_comb, float **z_comb, int *nlev_comb);

static double vapor_pressure (double t);

static int read_molecular_absorption (char *filename, 
				      float **zd, int *nlyr,
				      int quiet,
				      double **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic);

static int read_molecular_absorption_z (char* filename, int quiet,
					float **zd, int *nlyr);

static int column (float *dens, float *zd, int nlev, float altitude, float *column, int interpol_method, float *dens_air);
static int scale_profile (float *dens, float *zd, int nlev, float newcol, float altitude, 
			  float *factor, int interpol_method, float *dens_air);

static int precip_water (float *dens, float *zd, int nlev, float altitude, float *precip, int interpol_method, float *dens_air);
static int scale_precip (float *dens, float *zd, int nlev, float precip, float altitude, float *factor, int interpol_method, float *dens_air);

static int read_nc_atmosphere_grid (float **zd, int *nlev, char *filename, int quiet);


/**************************************************************/
/* Setup model atmosphere.                                    */
/**************************************************************/

int setup_atmosphere (input_struct input, output_struct *output)
{
  int i=0, status=0, lc=0, iv=0, iz=0, ialt=0, cdf=0;
  float scale_factor=0;
  float *zuser=NULL;
  float *zd_user=NULL;
  int   nlyr_user=0;
  int nuser=0;

  char filename[FILENAME_MAX];

  float *z_all=NULL;
  int nlev_all=-666;

  float *tmp_lambda=NULL, *tmp_dtauf=NULL, *tmp_dtauf_t=NULL;
  float **tmp_dtau=NULL;
  int tmp_nlambda=0;

  float tmpdens=0, idealdiff=0;
  int idealgas=0, firstnonideal=0;
  
  double boltzmann = 1.38065e-23;

  int monochromatic=0;

  int nlev_atm;
  float *dens_air_atm=NULL;
  float *z_atm=NULL;

  /* where to get the atmosphere from */
  if (input.ck_scheme == CK_GENERIC) {
    strcpy (filename, input.filename[FN_CK_GENERIC]);
    cdf=1;

    if (!input.quiet)
      if (strlen(input.filename[FN_ATMOSPHERE]) > 0)
	fprintf (stderr, " ... ignoring %s, reading atmosphere from %s\n",
		 input.filename[FN_ATMOSPHERE], input.filename[FN_CK_GENERIC]);
    
    /* determine altitude grid */
    status = read_nc_atmosphere_grid (&(output->atm.zd), &(output->atm.nlev), filename, input.quiet);
    if (status!=0) {
      fprintf (stderr, "Error %d returned by read_nc_atmosphere_grid()\n", status);
      return status;
    }
  }
  else {
    strcpy (filename, input.filename[FN_ATMOSPHERE]);
    cdf=0;

    /* determine altitude grid */
    status = read_1c_file_float (filename, &(output->atm.zd), &(output->atm.nlev));
    if (status!=0) {
      fprintf (stderr, "Error %d determining altitude profile from %s\n", status, filename);
      return status;
    }
  }


  /****************************/
  /* read atmospheric profile */
  /****************************/

  status = read_atmosphere (filename, &(output->atm), 
                            input.atm.interpol_method_press, 
                            input.atm.interpol_method_temper, 
                            input.atm.interpol_method_gas,
                            cdf, input.quiet, input.verbose);

  if (status!=0) {
    fprintf (stderr, "Error %d reading atmosphere from %s\n",
	     status, filename);
    return status;
  }

  /* copy air density on the grid of the atmosphere-file */
  nlev_atm = output->atm.nlev;

  /* copy the z-grid of the atmosphere file */
  z_atm = (float *) calloc ( nlev_atm, sizeof(float));
  for (lc=0; lc<output->atm.nlev; lc++)
    z_atm[lc] = output->atm.zd[lc];

  /* copy the air number density of the atmosphere file for linmix interpolations*/
  dens_air_atm =    calloc ( nlev_atm, sizeof(float));
  for (lc=0; lc<output->atm.nlev; lc++)
    dens_air_atm[lc] = output->atm.microphys.dens[MOL_AIR][lc];
   
  /* initialisation of the common z-grid: z_all */
  nlev_all = 0; 
  set_common_z_grid (output->atm.zd, nlev_atm, 
		     &(z_all), &(nlev_all));

  /********************************************************************************/
  /* combine all levels from other additional input files to a common grid: z_all */ 
  /********************************************************************************/

  /* dens                  */
  /* rh                    */
  /* refind                */
  /* rayscat               */
  /* molabs                */
  /* zout                  */
  /* altitude 2nd argument */
  /* altitude 1st argument */

  /* densities */
  for (i=0; i<MOL_NN; i++)
    if (strlen(input.atm.filename[i]) > 0)
      status += add_file2grid (&(z_all), &(nlev_all),
			       input.atm.filename[i]);

  /* relative humidity */
  if (strlen(input.filename[FN_RH]) > 0)
    status += add_file2grid (&(z_all), &(nlev_all),
			     input.filename[FN_RH]);

  /* refractive index */
  if (strlen(input.filename[FN_REFIND]) > 0)
    status += add_file2grid (&(z_all), &(nlev_all), 
			     input.filename[FN_REFIND]);


  /* rayleigh scattering cross section profile, user-defined */
  if (strlen(input.filename[FN_RAYLEIGH]) > 0) {
    status = read_molecular_absorption_z (input.filename[FN_RAYLEIGH], input.quiet,
					  &zd_user, &nlyr_user);
    if (status!=0) {
      fprintf (stderr, "Error reading z levels from user-defined Rayleigh scattering file %s\n",
	       input.filename[FN_RAYLEIGH]);
      fprintf (stderr, "Please check format and retry!\n");
      return -1;
    }

    /* add new levels to the common atmospheric profile */
    set_common_z_grid (zd_user, nlyr_user+1, &z_all, &nlev_all);

    /* free memory */
    free (zd_user);
  }


  /* molecular absorption cross section profile, user-defined  */
  if (strlen(input.filename[FN_MOLABS]) > 0) {
    status = read_molecular_absorption_z (input.filename[FN_MOLABS], input.quiet,
					  &zd_user, &nlyr_user);
    if (status!=0) {
      fprintf (stderr, "Error reading z levels from user-defined molecular absorption file %s\n", 
	       input.filename[FN_MOLABS]);
      fprintf (stderr, "Please check format and retry!\n");
      return -1;
    }

    /* add new levels to the common atmospheric profile */
    set_common_z_grid (zd_user, nlyr_user+1, &z_all, &nlev_all);

    /* free memory */
    free (zd_user);
  }


  /*  adding zout-levels to common grid , if zout_interpolate  */ 
  /*  if zout_interpolate is not activated zout-levels will    */
  /*  be added after the interpolation before redistribution   */ 
  if (input.atm.zout_interpolate == ZOUT_INTERPOLATE){

    /* user-defined equidistant altitude grid */
    if (input.atm.altitude_dz > 0) {
      nuser = (int) ((z_all[0] - z_all[nlev_all-1]) / 
		   input.atm.altitude_dz + 0.5) + 1;
    
      zuser = (float *) calloc (nuser, sizeof(float));

      for (i=0; i<nuser; i++) 
        zuser[i] = z_all[nlev_all-1] + input.atm.altitude_dz * (float) i;
    
      while (zuser[nuser-1] > z_all[0])
        nuser--;

      /* add user-defined grid to atmospheric grid */
      set_common_z_grid (zuser, nuser, &z_all, &nlev_all);

      free (zuser);
    }

    /* calculate output levels relative to sea level   */
    /* (user-defined data are relative to the surface) */
    output->atm.nzout    = input.atm.nzout;
    output->atm.zout_sur = calloc (output->atm.nzout, sizeof (float));
    output->atm.zout_sea = calloc (output->atm.nzout, sizeof (float));
  
    for (iz=0; iz<output->atm.nzout; iz++) {
      output->atm.zout_sur[iz] = input.atm.zout_sur[iz];
      
      if (input.atm.zout_sur[iz] == -999)
	output->atm.zout_sea[iz] = -999;
      else {
	output->atm.zout_sea[iz] = input.atm.zout_sur[iz] + output->atm.altitude;

        /* add new levels to the common atmospheric profile */
	set_common_z_grid (&(output->atm.zout_sea[iz]), 1, &(z_all), &(nlev_all));
      }
    }
  }

  /* add user-defined (ground) altitude as level */
  if (input.atm.altitude > 0) {
    zuser = (float *) calloc (1, sizeof(float));
    zuser[0] = input.atm.altitude;
    output->atm.altitude = input.atm.altitude;
    
    set_common_z_grid (zuser, 1, &(z_all), &nlev_all);
    free (zuser);
  }
  else
    output->atm.altitude = z_all[nlev_all-1];

  if (status!=0) {
    fprintf (stderr, "Error creating common atmospheric grid\n");
    return status;
  }

  /* for (lc=0; lc<nlev_all; lc++) */
  /*  fprintf (stderr, "z[%d] = %f\n", lc, z_all[lc]); */

  /************************************************************************/
  /* READING ADDITIONAL INPUT FILES                                       */ 
  /* INTERPOLATION TO THE COMMON Z-GRID: z_all                            */
  /* if input only covers a part of the vertical range of atmosphere_file */
  /* information of the atmosphere_file is assumed in the regions,        */
  /* which are not covered by the additional input file                   */
  /************************************************************************/

  /* pressure    */
  /* temperature */
  /* dens (dens_files might also be dens_tab files) */
  /* rh          */
  /* refind      */

  /* Pressure */
  status=0;
  status += interpolate_profile (z_atm, &(output->atm.microphys.press), output->atm.nlev, 
                                 z_all, nlev_all, input.atm.interpol_method_press,input.quiet);

  /* Temperature */
  status += interpolate_profile (z_atm, &(output->atm.microphys.temper), output->atm.nlev, 
                                 z_all, nlev_all, input.atm.interpol_method_temper,input.quiet);

/*   /\* this is not 100% consistent with interpolation of p and T *\/  */
  /* Air */
  status += interpolate_profile (z_atm, &(output->atm.microphys.dens[MOL_AIR]), output->atm.nlev, 
                                 z_all, nlev_all, input.atm.interpol_method_gas[MOL_AIR],input.quiet );

/*   /\* determine air number density from pressure and temperature *\/ */
/*   free(output->atm.microphys.dens[MOL_AIR]); */
/*   output->atm.microphys.dens[MOL_AIR] = calloc (nlev_all, sizeof(float)); */
/*   for (lc=0; lc<nlev_all; lc++) */
/*     output->atm.microphys.dens[MOL_AIR][lc] =  output->atm.microphys.press[lc] /  */
/*                                                (boltzmann * output->atm.microphys.temper[lc]) * 100.0 / 1e6; */

 
  /*********************/
  /* density profiles  */
  /*********************/

  for (i=0; i<MOL_NN; i++) {
      
    if ( i != MOL_AIR ) {  
      status = read_and_combine_density (input.atm.filename[i],
					 &output->atm, i, input.atm.interpol_method_gas,
					 output->atm.microphys.dens[MOL_AIR], z_all, nlev_all,
                                         dens_air_atm, z_atm, nlev_atm,
					 input.rte.solver, input.quiet, input.verbose);
      if (status !=0) {
	fprintf (stderr, "Error %d reading density profile\n", status);
        return status;
      }
    }
  }

  /*********************/
  /* relative humidity */
  /*********************/

  /* read_and_convert_rh() needs to be called in any case */
  /* because it calculates the otherwise empty rh profile */
  /* even if it doesn't read rh from the file             */
  status = read_and_convert_rh (input.filename[FN_RH], &(output->atm.microphys.rh),
                                z_all, nlev_all,
                                input.atm.interpol_method_rh,
				output->atm.microphys.temper,
				output->atm.microphys.dens[MOL_H2O],
                                input.quiet);
  if (status!=0) {
    fprintf (stderr, "Error %d returned by read_and_convert_rh()\n", status);
    return status;
  }


  /*********************/
  /* refraction index  */
  /*********************/

  /* calculate_or_read_refind() needs to be called in any case */
  /* because it calculates the otherwise empty refind profile  */
  /* even if it doesn't read refind from the file              */ 

  status = calculate_or_read_refind (input.filename[FN_REFIND], &(output->atm.microphys.refind),
                                     z_all, nlev_all,
                                     input.atm.interpol_method_refind, 
                                     output->atm.microphys.temper,
	                             output->atm.microphys.press,
                                     input.quiet);

  if (status!=0) {
    fprintf (stderr, "Error %d returned by calculate_or_read_refind() in atmosphere.c\n", status);
    return status;
  }


  /***********************************************/
  /*   copy altitude grid to final destination   */
  /***********************************************/

  free (output->atm.zd);
  output->atm.zd = calloc (nlev_all, sizeof(float));

  for (lc=0; lc<nlev_all; lc++)
    output->atm.zd[lc] = z_all[lc]; 

  output->atm.nlev = nlev_all;
  output->atm.nlyr = nlev_all-1; 

  free (z_all); 

  /* Test if surface altitude is really in the zd-grid */
  for (ialt=0; ialt<output->atm.nlev; ialt++) 
    if (output->atm.altitude == output->atm.zd[ialt]) 
      break;
    
  if (ialt == output->atm.nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid.\n", 
	     output->atm.altitude);
    return -1;
  }

  /*****************************************************************************************/
  /*                               Scaling profiles                                        */
  /*****************************************************************************************/ 

  /* Scale profiles of pressure and air density */
  if (input.pressure >= 0.0) {

    scale_factor = input.pressure / output->atm.microphys.press[ialt];
    for (lc=0; lc<output->atm.nlev; lc++) {
      output->atm.microphys.press[lc]         *= scale_factor;
      output->atm.microphys.dens[MOL_AIR][lc] *= scale_factor;
    } 
    
    if (input.verbose) {
      fprintf (stderr, " ... scaling pressure at %f km to %7.2f mbar\n", 
	       output->atm.zd[ialt], output->atm.microphys.press[ialt]);
      fprintf (stderr, " ... scaling density at %f km to %.6e\n", 
	       output->atm.zd[ialt], output->atm.microphys.dens[MOL_AIR][ialt]);
    }
  }

  /*****************************************************************************************/
  /*                      Scaling profiles with column values                              */
  /*****************************************************************************************/ 

  /* ????? Several things are missing here: Columns of O2, CO2,        ????? */
  /* ????? etc. may be defined via dens_column ... but the values      ????? */
  /* ????? defined there are not applied here                          ????? */

  /**********************/
  /* Scale ozone column */
  /**********************/

  status = column (output->atm.microphys.dens[MOL_O3], output->atm.zd, output->atm.nlev, 
                   output->atm.altitude, &(output->ozone_column),
                   input.atm.interpol_method_gas[MOL_O3], output->atm.microphys.dens[MOL_AIR]);

  if (status!=0) {
    fprintf (stderr, "Error %d calculating ozone column\n", status);
    return status;
  }

  if (input.atm.column[MOL_O3] >= 0.0) {

    switch (input.atm.unit[MOL_O3]) {
    case DU:
      /* Everything ok, do nothing */
      if (!input.quiet)
        fprintf (stderr, " ... scaling ozon column from %10.3f DU to ", 
	         output->ozone_column);
      break;
    case CM_2:
      if (!input.quiet)
        fprintf (stderr, " ... scaling ozon column from %e cm^-2 to ", 
	         output->ozone_column*2.687E16);
      input.atm.column[MOL_O3] /= 2.687E16;
      break;
    default:
      fprintf (stderr, "Error, unknown unit for O3 column");
      return -1;
    }

    scale_profile (output->atm.microphys.dens[MOL_O3], 
		   output->atm.zd, output->atm.nlev, 
		   input.atm.column[MOL_O3], output->atm.altitude,
		   &(output->ozone_scale_factor),
                   input.atm.interpol_method_gas[MOL_O3], output->atm.microphys.dens[MOL_AIR]);
    
    status = column (output->atm.microphys.dens[MOL_O3], output->atm.zd, output->atm.nlev, 
                     output->atm.altitude, &(output->ozone_column),
                     input.atm.interpol_method_gas[MOL_O3], output->atm.microphys.dens[MOL_AIR]); 

    if (status!=0) {
      fprintf (stderr, "\n Error %d calculating ozone column!\n", status);
      return status;
    }

    switch (input.atm.unit[MOL_O3]) {
    case DU:
      if (!input.quiet) fprintf (stderr, "%10.3f DU\n", output->ozone_column);
      break;
    case CM_2:
      if (!input.quiet) fprintf (stderr, "%e cm^-2\n", output->ozone_column*2.687E16);
      break;
    default:
      fprintf (stderr, "\n Error, unknown unit for O3 column");
      return -1;
    }
  }

  /****************************/
  /* Scale precipitable water */
  /****************************/
  status = precip_water (output->atm.microphys.dens[MOL_H2O], output->atm.zd, output->atm.nlev,
			 output->atm.altitude, &(output->precip_water),
                         input.atm.interpol_method_gas[MOL_H2O], output->atm.microphys.dens[MOL_AIR]);
  
  if (status!=0) {
    fprintf (stderr, "Error %d calculating precipitable water\n", status);
    return status;
  }
  
  if (input.precip_water >= 0.0) {

    if (!input.quiet)
      fprintf (stderr, " ... scaling precipitable water from %e kg/m2 to ", 
	       output->precip_water);
    
    status = scale_precip (output->atm.microphys.dens[MOL_H2O], 
			   output->atm.zd, output->atm.nlev, 
			   input.precip_water, output->atm.altitude,
			   &(output->water_scale_factor),
                           input.atm.interpol_method_gas[MOL_H2O], output->atm.microphys.dens[MOL_AIR]);
    
    if (status!=0) {
      fprintf (stderr, "Error %d returned by scale_precip()\n", status);
      return status;
    }

    status = precip_water (output->atm.microphys.dens[MOL_H2O], output->atm.zd, output->atm.nlev, 
			   output->atm.altitude, &(output->precip_water),
                           input.atm.interpol_method_gas[MOL_H2O], output->atm.microphys.dens[MOL_AIR]); 

    if (status!=0) {
      fprintf (stderr, "Error %d returned by precip_water()\n", status);
      return status;
    }

    if (!input.quiet)
      fprintf (stderr, "%e kg/m2\n", output->precip_water);
  }


  /********************/
  /* Scale NO2 column */
  /********************/

  status = column (output->atm.microphys.dens[MOL_NO2], output->atm.zd, output->atm.nlev,
                   output->atm.altitude, &(output->no2_column),
                   input.atm.interpol_method_gas[MOL_NO2], output->atm.microphys.dens[MOL_AIR]);
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }

  if (input.atm.column[MOL_NO2] >= 0.0) {
    switch (input.atm.unit[MOL_NO2]) {
    case DU:
      /* Everything ok, do nothing */
      break;
    case CM_2:
      input.atm.column[MOL_NO2] /= 2.687E16;
      break;
    default:
      fprintf (stderr, "Error, unknown unit for NO2 column");
      return -1;
    }

    scale_profile (output->atm.microphys.dens[MOL_NO2],
		   output->atm.zd, output->atm.nlev,
		   input.atm.column[MOL_NO2], output->atm.altitude,
		   &(output->no2_scale_factor),
                   input.atm.interpol_method_gas[MOL_NO2], output->atm.microphys.dens[MOL_AIR]);
    
    status = column (output->atm.microphys.dens[MOL_NO2], output->atm.zd, output->atm.nlev,
                     output->atm.altitude, &(output->no2_column),
                     input.atm.interpol_method_gas[MOL_NO2], output->atm.microphys.dens[MOL_AIR]);
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor */
    /* and the name is misleading, however, this is only useful for airmass   */
    /* factor calculations for which the column is 0.0 when no absorber       */
    /* is present                                                             */
    if (output->atm.microphys.denstab_id == MOL_NO2) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_NO2];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_NO2];
	}
      }
    }
  }

  /********************/
  /* Scale BrO column */
  /********************/
  status = column (output->atm.microphys.dens[MOL_BRO], output->atm.zd, output->atm.nlev,
                   output->atm.altitude, &(output->bro_column),
                   input.atm.interpol_method_gas[MOL_BRO], output->atm.microphys.dens[MOL_AIR]);
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }
  
  if ( input.atm.column[MOL_BRO] >= 0.0 ) {
    switch (input.atm.unit[MOL_BRO]) {
    case DU:
      /* Everything ok, do nothing */
      break;
    case CM_2:
      input.atm.column[MOL_BRO] /= 2.687E16;
      break;
    default:
      fprintf (stderr, "Error, unknown unit for BrO column");
      return -1;
    }

    scale_profile (output->atm.microphys.dens[MOL_BRO],
		   output->atm.zd, output->atm.nlev,
		   input.atm.column[MOL_BRO], output->atm.altitude,
		   &(output->bro_scale_factor),
                   input.atm.interpol_method_gas[MOL_BRO], output->atm.microphys.dens[MOL_AIR]);
    
    status = column (output->atm.microphys.dens[MOL_BRO], output->atm.zd, output->atm.nlev,
                     output->atm.altitude, &(output->bro_column),
                     input.atm.interpol_method_gas[MOL_BRO], output->atm.microphys.dens[MOL_AIR]);
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor      */
    /* and the name is misleading, however, this is only useful for airmass        */
    /* factor calculations for which the column is 0.0 when no absorber is present */
    if (output->atm.microphys.denstab_id == MOL_BRO) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_BRO];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_BRO];
	}
      }
    }
  }

  /*********************/
  /* Scale OClO column */
  /*********************/
  status = column (output->atm.microphys.dens[MOL_OCLO], output->atm.zd, output->atm.nlev,
                   output->atm.altitude, &(output->oclo_column),
                   input.atm.interpol_method_gas[MOL_OCLO], output->atm.microphys.dens[MOL_AIR]);
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }
  
  if ( input.atm.column[MOL_OCLO] >= 0.0 ) {
    switch (input.atm.unit[MOL_OCLO]) {
    case DU:
      /* Everything ok, do nothing */
      break;
    case CM_2:
      input.atm.column[MOL_OCLO] /= 2.687E16;
      break;
    default:
      fprintf (stderr, "Error, unknown unit for OClO column");
      return -1;
    }

    scale_profile (output->atm.microphys.dens[MOL_OCLO],
		   output->atm.zd, output->atm.nlev,
		   input.atm.column[MOL_OCLO], output->atm.altitude,
		   &(output->oclo_scale_factor),
                   input.atm.interpol_method_gas[MOL_OCLO], output->atm.microphys.dens[MOL_AIR]);
    
    status = column (output->atm.microphys.dens[MOL_OCLO], output->atm.zd, output->atm.nlev,
                     output->atm.altitude, &(output->oclo_column),
                     input.atm.interpol_method_gas[MOL_OCLO], output->atm.microphys.dens[MOL_AIR]);
    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor      */
    /* and the name is misleading, however, this is only useful for airmass        */
    /* factor calculations for which the column is 0.0 when no absorber is present */
    if (output->atm.microphys.denstab_id == MOL_OCLO) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_OCLO];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_OCLO];
	}
      }
    }
  }

  /*********************/
  /* Scale HCHO column */
  /*********************/
  status = column (output->atm.microphys.dens[MOL_HCHO], output->atm.zd, output->atm.nlev,
                   output->atm.altitude, &(output->hcho_column),
                   input.atm.interpol_method_gas[MOL_HCHO], output->atm.microphys.dens[MOL_AIR] );
  if (status!=0) {
    fprintf (stderr, "Error %d calculating column\n", status);
    return status;
  }
  
  if ( input.atm.column[MOL_HCHO] >= 0.0 ) {
    switch (input.atm.unit[MOL_HCHO]) {
    case DU:
      /* Everything ok, do nothing */
      break;
    case CM_2:
      input.atm.column[MOL_HCHO] /= 2.687E16;
      break;
    default:
      fprintf (stderr, "Error, unknown unit for HCHO column");
      return -1;
    }

    scale_profile (output->atm.microphys.dens[MOL_HCHO],
		   output->atm.zd, output->atm.nlev,
		   input.atm.column[MOL_HCHO], output->atm.altitude,
		   &(output->hcho_scale_factor),
                   input.atm.interpol_method_gas[MOL_HCHO], output->atm.microphys.dens[MOL_AIR] );
    
    status = column (output->atm.microphys.dens[MOL_HCHO], output->atm.zd, output->atm.nlev,
                     output->atm.altitude, &(output->hcho_column),
                     input.atm.interpol_method_gas[MOL_HCHO], output->atm.microphys.dens[MOL_AIR]);

    if (status!=0) {
      fprintf (stderr, "Error %d calculating column\n", status);
      return status;
    }

    /* If density file multiply with column value, hence it is a scale factor      */
    /* and the name is misleading, however, this is only useful for airmass        */
    /* factor calculations for which the column is 0.0 when no absorber is present */
    if (output->atm.microphys.denstab_id == MOL_HCHO) {
      for (i=1; i<=output->atm.microphys.nsza_denstab; i++) {
	for (lc=0; lc<output->atm.nlev; lc++) {
	  output->atm.microphys.denstab[output->atm.microphys.denstab_id][i-1][lc] *= input.atm.column[MOL_HCHO];
	  output->atm.microphys.denstab_amf[i-1][lc]                     *= input.atm.column[MOL_HCHO];
	}
      }
    }
  }


  /*****************************************************************************************/
  /*                      Scaling profiles with mixing ratios                              */
  /*                         at the user-defined altitude                                  */
  /*****************************************************************************************/ 

  for (ialt=0; ialt<output->atm.nlev; ialt++)
    if (output->atm.altitude == output->atm.zd[ialt])
      break;
    
  if (ialt == output->atm.nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid.\n", 
	     output->atm.altitude);
    return -1;
  }

  output->mixing_ratio = calloc (MX_NN, sizeof(float));

  /* O2 */
  if (input.mixing_ratio[MX_O2]<0) {
    /* If not specified, use the mixing ratio at the user-defined altitude */
    output->mixing_ratio[MX_O2] = output->atm.microphys.dens[MOL_O2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_O2] = input.mixing_ratio[MX_O2];
    
    scale_factor = (input.mixing_ratio[MX_O2] / 1e6) / 
      (output->atm.microphys.dens[MOL_O2] [ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling O2-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_O2][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	       input.mixing_ratio[MX_O2]); 

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_O2][lc] *= scale_factor;
  }
  

  /* CO2 */
  if (input.mixing_ratio[MX_CO2]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_CO2] = output->atm.microphys.dens[MOL_CO2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_CO2] = input.mixing_ratio[MX_CO2];
    
    scale_factor = (input.mixing_ratio[MX_CO2] / 1e6) / 
      (output->atm.microphys.dens[MOL_CO2][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling CO2-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_CO2][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	      input.mixing_ratio[MX_CO2]); 

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_CO2][lc] *= scale_factor;
  }
  

  /* H2O */
  if (input.mixing_ratio[MX_H2O]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_H2O] = output->atm.microphys.dens[MOL_H2O][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_H2O] = input.mixing_ratio[MX_H2O];
    
    scale_factor = (input.mixing_ratio[MX_H2O] / 1e6) / 
      (output->atm.microphys.dens[MOL_H2O][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling H2O-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_H2O][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	      input.mixing_ratio[MX_H2O]); 

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_H2O][lc] *= scale_factor;
  }
  
  /* NO2 */
  if (input.mixing_ratio[MX_NO2]<0) {
    /* If not specified, use the mixing ratio of the lowest level in the */
    /* atmospheric data file.                                            */

    output->mixing_ratio[MX_NO2] = output->atm.microphys.dens[MOL_NO2][ialt] / 
      output->atm.microphys.dens[MOL_AIR][ialt] * 1e6;
  }
  else {
    output->mixing_ratio[MX_NO2] = input.mixing_ratio[MX_NO2];
    
    scale_factor = (input.mixing_ratio[MX_NO2] / 1e6) / 
      (output->atm.microphys.dens[MOL_NO2][ialt] / 
       output->atm.microphys.dens[MOL_AIR][ialt]);

    if (!input.quiet)
      fprintf(stderr," ... scaling NO2-mixing ratio from %10.5e ppm to %10.5e ppm\n",
	      (output->atm.microphys.dens[MOL_NO2][ialt]/output->atm.microphys.dens[MOL_AIR][ialt])*1e6,
	      input.mixing_ratio[MX_NO2]);

    for (lc=0; lc<output->atm.nlev; lc++)
      output->atm.microphys.dens[MOL_NO2][lc] *= scale_factor;
  }
  
  /* default CH4, N2O, F11, F12, and F22 are 'hardwired'; */
  /* these gases are assumed to be well-mixed throughout  */
  /* the whole atmosphere                                 */

  if (input.mixing_ratio[MX_CH4] < 0)
    output->mixing_ratio[MX_CH4] = 1.6; 
  else {
    output->mixing_ratio[MX_CH4] = input.mixing_ratio[MX_CH4];
    if (!input.quiet)
      fprintf(stderr," ... scaling CH4-mixing ratio from 1.6 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_CH4]);
  }

  if (input.mixing_ratio[MX_N2O] < 0)
    output->mixing_ratio[MX_N2O] = 0.28; 
  else {
    output->mixing_ratio[MX_N2O] = input.mixing_ratio[MX_N2O];
    if (!input.quiet)
      fprintf(stderr," ... scaling N2O-mixing ratio from 0.28 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_N2O]);
  }
  
  if (input.mixing_ratio[MX_F11] < 0)
    output->mixing_ratio[MX_F11] = 0.268E-3; 
  else {
    output->mixing_ratio[MX_F11] = input.mixing_ratio[MX_F11];
    if (!input.quiet)
      fprintf(stderr," ... scaling F11-mixing ratio from 0.268E-3 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_F11]);
  }
  
  if (input.mixing_ratio[MX_F12] < 0)
    output->mixing_ratio[MX_F12] = 0.503E-3; 
  else {
    output->mixing_ratio[MX_F12] = input.mixing_ratio[MX_F12];
    if (!input.quiet)
      fprintf(stderr," ... scaling F12-mixing ratio from 0.503E-3 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_F12]);   
  }

  if (input.mixing_ratio[MX_F22] < 0)
    output->mixing_ratio[MX_F22] = 0.105E-3; 
  else {
    output->mixing_ratio[MX_F22] = input.mixing_ratio[MX_F22];
    if (!input.quiet)
      fprintf(stderr," ... scaling F22-mixing ratio from 0.105E-3 ppm to %10.5e ppm\n",
	      input.mixing_ratio[MX_F22]);
  }

  /* Read Rayleigh cross section if specified */
  if (strlen(input.filename[FN_RAYLEIGH]) > 0) {
    status = read_molecular_absorption (input.filename[FN_RAYLEIGH],
					&zd_user, &nlyr_user,
					input.quiet,
					&(output->atm.optprop.tau_rayleigh_user), 
					&tmp_dtau, &tmp_lambda, &tmp_nlambda,
					&monochromatic);

    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_molecular_absorption (\"%s\")\n", 
	       status, input.filename[FN_RAYLEIGH]);
      return status;
    }

    if (monochromatic) {
      output->atm.rayleigh = RAYLEIGH_FILE_MONO;

      if (!input.quiet)
	fprintf (stderr, " ... read monochromatic molecular absorption from %s\n", 
		 input.filename[FN_RAYLEIGH]);
    }
    else {
      output->atm.rayleigh = RAYLEIGH_FILE_SPEC;
      
      if (!input.quiet)
	fprintf (stderr, " ... read Rayleigh scattering cross section from %s\n", 
		 input.filename[FN_RAYLEIGH]);
      
      /***************************************************************/
      /* interpolate optical thickness to transmittance grid, linear */
      /***************************************************************/

      /* allocate memory for optical thickness on internal transmittance grid */
      output->atm.optprop.tau_rayleigh_user_t = calloc (output->wl.nlambda_t, sizeof(double *));
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	output->atm.optprop.tau_rayleigh_user_t[iv] = calloc (nlyr_user, sizeof(double));
      
      switch (output->wl.type) {
      case WLGRID_CK:
      case WLGRID_USER:
      case WLGRID_BANDS:
      case WLGRID_UVSPEC:
      case WLGRID_MOLABS:
	
	/* memory for temporary arrays */
	tmp_dtauf   = calloc (tmp_nlambda,          sizeof (float));
	tmp_dtauf_t = calloc (output->wl.nlambda_t, sizeof (float));
	
	/* loop over layers */
	for (lc=0; lc<nlyr_user; lc++) {
	  
	  /* interpolate tmp_dtauf to internal wavelength grid */

	  /* copy double array to temporary float array */
	  /* ??? should improve that by providing a ??? */
	  /* ??? an arb_wvn for arrays of double    ??? */
	  for (iv=0; iv<tmp_nlambda; iv++)
	    tmp_dtauf[iv] = tmp_dtau[lc][iv];
	  
	  status = arb_wvn (tmp_nlambda, tmp_lambda, tmp_dtauf, 
			    output->wl.nlambda_t, output->wl.lambda_t, 
			    tmp_dtauf_t,
			    1, 0);
	  
	  /* copy temporary float array to final destination */
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	    output->atm.optprop.tau_rayleigh_user_t[iv][lc] = tmp_dtauf_t[iv];
	  
	  
	  if (status != 0) {
	    fprintf (stderr, "Error %d interpolating molecular absorption grid to internal\n", status);
	    fprintf (stderr, "wavelength grid\n");
	    return status;
	  }
	}

	free(tmp_dtauf); free(tmp_dtauf_t);
	break;
	  
      default:
	fprintf (stderr, "Error, unknown wavelength grid type %d\n", 
		 output->wl.type);
	return -1;
      }

      free(tmp_lambda); 
      for (lc=0; lc<nlyr_user; lc++)
	free (tmp_dtau[lc]);
      free (tmp_dtau);
    }


    /* redistribute tau_rayleigh_user and tau_rayleigh_user_t to common atmospheric grid */
    status = 0;
    if (monochromatic) {
      status += redistribute_1D ((void **) (&(output->atm.optprop.tau_rayleigh_user)),
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1, 
				 1, REDISTRIBUTE_DOUBLE);
    }
    else {
      status += redistribute_2D ((void ***) (&(output->atm.optprop.tau_rayleigh_user_t)), 
				 output->wl.nlambda_t, 
				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1,
				 1, REDISTRIBUTE_DOUBLE);
    }

    if (status!=0) {
      fprintf (stderr, "Error interpolating user-defined molecular absorption coefficients\n");
      fprintf (stderr, "to common atmospheric vertical grid.\n");
      return status;
    }
  }			



  /* Read molecular absorption cross section if specified */
  if (strlen(input.filename[FN_MOLABS]) > 0) {
    status = read_molecular_absorption (input.filename[FN_MOLABS], 
					&zd_user, &nlyr_user,
					input.quiet,
					&(output->atm.optprop.tau_molabs_user), 
					&tmp_dtau, &tmp_lambda, &tmp_nlambda,
					&monochromatic);

    if (status!=0) {
      fprintf (stderr, "Error, status %d returned by read_molecular_absorption (\"%s\")\n", 
	       status, input.filename[FN_MOLABS]);
      return status;
    }

    if (monochromatic) {
      output->atm.molabs = MOLABS_FILE_MONO;

      if (!input.quiet)
	fprintf (stderr, " ... read monochromatic molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
    }
    else {
      output->atm.molabs = MOLABS_FILE_SPEC;
      
      if (!input.quiet)
	fprintf (stderr, " ... read spectral molecular absorption from %s\n", 
		 input.filename[FN_MOLABS]);
      
      /***************************************************************/
      /* interpolate optical thickness to transmittance grid, linear */
      /***************************************************************/

      /* allocate memory for optical thickness on internal transmittance grid */
      output->atm.optprop.tau_molabs_user_t = calloc (output->wl.nlambda_t, sizeof(double *));
      for (iv=0; iv<output->wl.nlambda_t; iv++)
	output->atm.optprop.tau_molabs_user_t[iv] = calloc (nlyr_user, sizeof(double));
      
      switch (output->wl.type) {
      case WLGRID_CK:
      case WLGRID_USER:
      case WLGRID_BANDS:
      case WLGRID_UVSPEC:
	
	/* memory for temporary arrays */
	tmp_dtauf   = calloc (tmp_nlambda,          sizeof (float));
	tmp_dtauf_t = calloc (output->wl.nlambda_t, sizeof (float));
	
	/* loop over layers */
	for (lc=0; lc<nlyr_user; lc++) {
	  
	  /* interpolate tmp_dtauf to internal wavelength grid */

	  /* copy double array to temporary float array */
	  /* ??? should improve that by providing a ??? */
	  /* ??? an arb_wvn for arrays of double    ??? */
	  for (iv=0; iv<tmp_nlambda; iv++)
	    tmp_dtauf[iv] = tmp_dtau[lc][iv];
	  
	  status = arb_wvn (tmp_nlambda, tmp_lambda, tmp_dtauf, 
			    output->wl.nlambda_t, output->wl.lambda_t, 
			    tmp_dtauf_t,
			    1, 0);
	  
	  /* copy temporary float array to final destination */
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	    output->atm.optprop.tau_molabs_user_t[iv][lc] = tmp_dtauf_t[iv];
	  
	  
	  if (status != 0) {
	    fprintf (stderr, "Error %d interpolating molecular absorption grid to internal\n", status);
	    fprintf (stderr, "wavelength grid\n");
	    return status;
	  }
	}

	free(tmp_dtauf); free(tmp_dtauf_t);
	break;
	  
      case WLGRID_MOLABS:
	/* no need to interpolate, just copy */

	/* copy temporary float array to final destination */
	for (lc=0; lc<nlyr_user; lc++)
	  for (iv=0; iv<output->wl.nlambda_t; iv++)
	    output->atm.optprop.tau_molabs_user_t[iv][lc] = tmp_dtau[lc][iv];
	
	break;
	
      default:
	fprintf (stderr, "Error, unknown wavelength grid type %d\n", 
		 output->wl.type);
	return -1;
      }

      free(tmp_lambda); 
      for (lc=0; lc<nlyr_user; lc++)
	free (tmp_dtau[lc]);
      free (tmp_dtau);
    }


    /* redistribute tau_molabs_user and tau_molabs_user_t to common atmospheric grid */
    status = 0;
    if (monochromatic) {
      status += redistribute_1D ((void **) (&(output->atm.optprop.tau_molabs_user)),
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1, 
				 1, REDISTRIBUTE_DOUBLE);
    }
    else {
      status += redistribute_2D ((void ***) (&(output->atm.optprop.tau_molabs_user_t)), 
				 output->wl.nlambda_t, 
				 output->wl.nlambda_rte_lower, output->wl.nlambda_rte_upper,
				 zd_user, nlyr_user, 
				 output->atm.zd, output->atm.nlev-1,
				 1, REDISTRIBUTE_DOUBLE);
    }

    if (status!=0) {
      fprintf (stderr, "Error interpolating user-defined molecular absorption coefficients\n");
      fprintf (stderr, "to common atmospheric vertical grid.\n");
      return status;
    }
  }			

  /* Set common vertical resolution */
  output->atm.nlev_common = 0;
  set_common_z_grid (output->atm.zd, output->atm.nlev,
		     &output->atm.zd_common, &output->atm.nlev_common);
  
  if (input.verbose) {

    fprintf (stderr, "*** setup_atmosphere()\n");
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
    fprintf (stderr, "  lc |  z[km] |  Pressure  | Temp.  |    Air      |   Ozone     |     O2      | Water vap.  |    CO2      |    NO2      | \n");
    fprintf (stderr, "     |        |   [mbar]   |  [K]   |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    |   [cm-3]    | \n");
    fprintf (stderr, "-------------------------------------------------------------------------------------------------------------------------------------\n");

    for (lc=0; lc<output->atm.nlev; lc++)
      fprintf (stderr, "%5d   %6.2f   %10.5f   %6.2f   %.5e   %.5e   %.5e   %.5e   %.5e   %.5e\n",
	       lc, output->atm.zd[lc],
	       output->atm.microphys.press[lc], output->atm.microphys.temper[lc],output->atm.microphys.dens[MOL_AIR][lc],
	       output->atm.microphys.dens[MOL_O3][lc],output->atm.microphys.dens[MOL_O2][lc], output->atm.microphys.dens[MOL_H2O][lc],
               output->atm.microphys.dens[MOL_CO2][lc],output->atm.microphys.dens[MOL_NO2][lc] );

    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");
    fprintf (stderr, "%5s | %6.2f | %10.5f | %6.2f |     %7.3f |  %7.3f DU |     %7.3f | %6.3f kg/m2 |    %7.3f | %7.5f DU |\n",
	     "sum", 0.0/0.0,
	     0.0/0.0, 0.0/0.0, 0.0/0.0,
	     output->ozone_column, 0.0/0.0, output->precip_water, 0.0/0.0, output->no2_column);
    fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n");

/*     fprintf (stderr, "*** setup_atmosphere()\n"); */
/*     fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n"); */
/*     fprintf (stderr, "   lc |  z[km] |  Pressure  | Temp.  |   Ozone     | Water vap. \n"); */
/*     fprintf (stderr, "      |        |   [mbar]   |  [K]   |   [cm-3]    |  [cm-3]    \n"); */
/*     fprintf (stderr, " -------------------------------------------------------------------------------------------------------------------------------------\n"); */

/*     for (lc=0; lc<output->atm.nlev; lc++) */
/*       fprintf (stderr, "%5d | %6.2f | %10.5f | %6.2f | %.5e | %.5e\n", */
/* 	       lc, output->atm.zd[lc], */
/* 	       output->atm.microphys.press[lc], output->atm.microphys.temper[lc], */
/* 	       output->atm.microphys.dens[MOL_O3][lc], output->atm.microphys.dens[MOL_H2O][lc]); */

  }

  /* if zout_interpolate is not specified, data for                  */
  /* altitude (2nd argument) levels will be added in redistribute.c  */
  /* zout-Levels will not be added anywhere to the z-grid            */
  /* solvers do that for us                                          */

  /* this cant be done before verbose output as data for equidistant grid would be missing */
  if (input.atm.zout_interpolate == NO_ZOUT_INTERPOLATE){

    /*  user-defined equidistant altitude grid  */
    if (input.atm.altitude_dz > 0) {
      nuser = (int) ((output->atm.zd[0] - output->atm.zd[output->atm.nlev-1]) / 
                      input.atm.altitude_dz + 0.5) + 1;
      
      zuser = (float *) calloc (nuser, sizeof(float));

      for (i=0; i<nuser; i++)
        zuser[i] = output->atm.zd[output->atm.nlev-1] + input.atm.altitude_dz * (float) i;
  
      while (zuser[nuser-1] > output->atm.zd[0])
        nuser--;

      /* add user-defined grid to atmospheric grid */
      set_common_z_grid (zuser, nuser, &output->atm.zd_common, &output->atm.nlev_common);

      free (zuser);
    }

    /* calculate output levels relative to sea level   */
    /* (user-defined data are relative to the surface) */
    
    output->atm.nzout    = input.atm.nzout;
    output->atm.zout_sur = calloc (output->atm.nzout, sizeof (float));
    output->atm.zout_sea = calloc (output->atm.nzout, sizeof (float));
    
    for (iz=0; iz<output->atm.nzout; iz++) {
      output->atm.zout_sur[iz] = input.atm.zout_sur[iz];
      
      if (input.atm.zout_sur[iz] == -999)
        output->atm.zout_sea[iz] = -999;
      else
        output->atm.zout_sea[iz] = input.atm.zout_sur[iz] + output->atm.altitude;
    }
    /* zout level are not added to any z-grid */
    /* when input.atm.zout_interpolate == NO_ZOUT_INTERPOLATE */
  }


  switch (input.processing){
    case PROCESS_NONE:
    case PROCESS_SUM:
    case PROCESS_INT:
      break;
    case PROCESS_HEATING:
    case PROCESS_HEATING_UL:
    case PROCESS_HEATING_LCL:
      output->atm.microphys.dens_air    = calloc (output->atm.nlev, sizeof (float));
      for (iz=0; iz<output->atm.nlev; iz++)
        output->atm.microphys.dens_air[iz] = output->atm.microphys.dens[MOL_AIR][iz];
        
      status = interpolate_profile (output->atm.zd, &output->atm.microphys.dens_air, output->atm.nlev,
                                    output->atm.zout_sur, output->atm.nzout, 
                                    input.atm.interpol_method_gas[MOL_AIR],input.quiet);
    
      if (status!=0) {
        fprintf (stderr, "Error interpolating dens_air\n");
        return status;
      }
      break;
    default:
      fprintf (stderr, "Error, unknown processing option in atmosphere.c\n");
      return -1;
  }

  /* this is just a check for stupid programming errors, in usual case this is unnessesary */

  /* Finally, check if profiles of pressure, temperature and air density are consistent */
  /* with ideal gas law over 6 orders of magnitude in pressure; check only first six    */
  /* orders of magnitude because for very small pressures roundoff errors due to too    */
  /* few digits cause problems                                                          */
  firstnonideal=1;
  for (lc=0; lc<output->atm.nlev; lc++) {

    idealgas=1;

    if (output->atm.microphys.press[lc]/output->atm.microphys.press[output->atm.nlev-1]>1e-6) {
      tmpdens = output->atm.microphys.press[lc]*1E-04 / (boltzmann*output->atm.microphys.temper[lc]);
      
      if ((tmpdens==0 && output->atm.microphys.dens[MOL_AIR][lc]!=0) || 
	  (tmpdens!=0 && output->atm.microphys.dens[MOL_AIR][lc]==0))
	idealgas=0;
      
      if (output->atm.microphys.dens[MOL_AIR][lc]>0)
	if ((idealdiff = fabs ((tmpdens - output->atm.microphys.dens[MOL_AIR][lc]) / output->atm.microphys.dens[MOL_AIR][lc])) > 0.01)
	  idealgas=0;
      
      if (!idealgas && firstnonideal) {
	firstnonideal=0;
	fprintf (stderr, "\n");
	fprintf (stderr, "*** ATTENTION: pressure, temperature, and air density not\n");
	fprintf (stderr, "*** consistent with ideal gas law at the following levels\n");
	fprintf (stderr, "*** (assuming a Boltzmann constant %.6e):\n", boltzmann);
	fprintf (stderr, "***    z [km]  density [cm-3]  p/kT [cm-3]  difference\n");
      }
      
      if (!idealgas) {
	fprintf (stderr, "***   %7.2f  %.4e      %.4e    %6.1f%%\n", 
		 output->atm.zd[lc], output->atm.microphys.dens[MOL_AIR][lc], tmpdens, idealdiff*100.0);
      }
    }
  }
  
  if (!firstnonideal) {
    fprintf (stderr, "*** Please check pressure, temperature, and air density in\n");
    fprintf (stderr, "*** %s for consistency!\n", input.filename[FN_ATMOSPHERE]);
    fprintf (stderr, "\n");
    return -1;
  } 

  return 0;
}


/**************************************************************/
/* Setup optical properties of trace gases.                   */
/**************************************************************/

int setup_gases (input_struct input, output_struct *output)
{
  static int first=1;
  int status=0;

  double bscar=0;
  double babso=0, babso_o3=0, babso_co2=0, babso_o2=0;
  double babso_no2=0, babso_bro=0, babso_oclo=0, babso_hcho=0, babso_md=0;
  double babso_amf=0, babso_amftot=0, vertical_column=0, crs_amf=0;
  double bscartot=0;
  double babsotot=0, babsotot_o3=0, babsotot_co2=0, babsotot_o2=0; 
  double babsotot_no2=0, babsotot_bro=0, babsotot_oclo=0, babsotot_hcho=0;
  double deltaz=0;
  int iq=0, is=0, isthis=0, iv=0, lc=0;
  int nq_rayleigh=0;
  int nlev = output->atm.nlev;
  int nlyr = output->atm.nlyr;

  int iq1=0, iq2=0, iq3=0, iq4=0;
  int nq=0, nq1=0, nq2=0, nq3=0, nq4=0;

  double factor=0;
  double weight[KATO_COMPONENTS];

  float *fu_rayleigh = calloc (nlyr, sizeof(float));
  float density=-999.;
  int gas_nr;

  /* number of quadrature points for Rayleigh scattering */
  switch (input.ck_scheme) {
  case CK_FU:
    nq_rayleigh=FU_MAXINT;
    break;
    
  case CK_KATO:
  case CK_KATO2:
  case CK_KATO2_96:
  case CK_AVHRR_KRATZ:
  case CK_GENERIC:
  case CK_LOWTRAN:
  case CK_NONE:
    nq_rayleigh=1;
    break;

  default:
    fprintf (stderr, "Error: unimplemented correlated-k scheme %d\n", input.ck_scheme);
    return -1;
    break;
  }

  if (first!=0) {
    first = 0;

    /* allocate memory for profiles */
    output->atm.cdento   = (float *) calloc (nlyr, sizeof(float));
  
    output->atm.wght_t   = calloc (output->wl.nlambda_t, sizeof (double *));
    output->atm.nq_t     = calloc (output->wl.nlambda_t, sizeof (int));

    output->mc_photons_t = calloc (output->wl.nlambda_t, sizeof (double));


    /* copy ck weights */
    for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
      
      switch (input.ck_scheme) {
      case CK_FU:
	output->atm.nq_t[iv] = output->ck.wght[FU_WGHT][iv+1];
	break;
	
      case CK_KATO:
      case CK_KATO2:
      case CK_KATO2_96:
	output->atm.nq_t[iv] = output->ck.wght[KATO_CO2][iv+1] *
                               output->ck.wght[KATO_O3] [iv+1] *
                               output->ck.wght[KATO_O2] [iv+1] *
                               output->ck.wght[KATO_H2O][iv+1];
	break;

      case CK_AVHRR_KRATZ:
	output->atm.nq_t[iv] = output->ck.wght[AVHRR_KRATZ_WGHT][iv+1];
	break;

      case CK_GENERIC:
	output->atm.nq_t[iv] = output->ck.wght[GENERIC_WGHT][iv+1];
	break;

      case CK_LOWTRAN:
	output->atm.nq_t[iv] = LOWTRAN_MAXINT;
	break;

      case CK_NONE:
	output->atm.nq_t[iv] = 1;
	break;
	
      default:
	fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
	return -1;
	break;
      }

      output->atm.wght_t [iv] = calloc (output->atm.nq_t[iv], sizeof (double ));

      /* allocate memory for photon fractions */
      output->mc_photons_t [iv] = calloc (output->atm.nq_t[iv], sizeof (double ));


      /* copy photon fractions */
      switch(input.ck_scheme) {
      case CK_NONE:
	
	iq=0; /* no loop over quadrature points required */
	output->mc_photons_t [iv][iq] = output->wl.pfraction[iv];
	
	break;
	
      case CK_KATO:
      case CK_KATO2:
      case CK_KATO2_96:
	nq1 = output->crs_ck.profile[KATO_CO2-1][iv].ngauss;
	nq2 = output->crs_ck.profile[KATO_O3-1 ][iv].ngauss;
	nq3 = output->crs_ck.profile[KATO_O2-1 ][iv].ngauss;
	nq4 = output->crs_ck.profile[KATO_H2O-1][iv].ngauss;
      
	iq=0;  /* reset quadrature counter */
	/* loop over quadrature points */
	for (iq1=0; iq1<nq1; iq1++)  /* CO2 */
	  for (iq2=0; iq2<nq2; iq2++)  /* O3 */
	    for (iq3=0; iq3<nq3; iq3++)  /* O2 */
	      for (iq4=0; iq4<nq4; iq4++)  /* H2O */
		switch (input.source) {
		case SRC_SOLAR: /* solar source */
		  output->mc_photons_t [iv][iq++] = ((double *****) output->ck.x_solar)  [iv+1][iq1+1][iq2+1][iq3+1][iq4+1];
		  break;
		case SRC_THERMAL: /* thermal source */
		  output->mc_photons_t [iv][iq++] = ((double *****) output->ck.x_thermal)[iv+1][iq1+1][iq2+1][iq3+1][iq4+1];
		  break;
		default:
		  fprintf (stderr, "Error, unknown source %d\n", input.source);
		  return -1;
		}
	break;
      
      case CK_FU:
      case CK_AVHRR_KRATZ:
      case CK_GENERIC:
      case CK_LOWTRAN:
	
	nq = output->crs_ck.profile[0][iv].ngauss;
	
	for (iq=0; iq<nq; iq++)
	  switch (input.source) {
	  case SRC_SOLAR: /* solar source */
	    output->mc_photons_t [iv][iq] = ((double **) output->ck.x_solar)  [iv+1][iq+1];
	    break;
	  case SRC_THERMAL: /* thermal source */
	    output->mc_photons_t [iv][iq] = ((double **) output->ck.x_thermal)[iv+1][iq+1];
	    break;
	  default:
	    fprintf (stderr, "Error, unknown source %d\n", input.source);
	    return -1;
	  }
	
	
	break;
	
      default:
	fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
	return -1;
	break;
      }
    }

    /* Rayleigh scattering optical thickness */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_rayleigh_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;

    /* Molecular absorption optical thickness */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_molabs_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;

    /* Molecular absorption optical thickness for AMF calculations */
    if ((status = ASCII_calloc_double_3D_arylen (&output->atm.optprop.tau_molabs_md_t, 
						 nlyr, output->wl.nlambda_t, 
						 output->atm.nq_t)) != 0)
      return status;
    
    /* Calculate column density of air (total density of the atmosphere). */
    /* Required to calculate Rayleigh optical depth.                      */

    if (output->atm.rayleigh==RAYLEIGH_CALC) {
      for (lc=0; lc<nlyr; lc++) {
	
	/* The factor 1e5 converts z from km to cm.*/
	deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;
      
	/* average air density over a layer assuming  dens = exp (-c * z) */
        output->atm.cdento[lc] = log_average(output->atm.microphys.dens[MOL_AIR][lc],output->atm.microphys.dens[MOL_AIR][lc+1]) * deltaz;
      }
    }
  } /* End if first block */

  if ( output->atm.microphys.denstab_id ) {
    /* Find matrix profile corresponding to the sza this rte calculation is done for */
    /* This is airmass factor stuff */
    iv=0;  /* No solar zenith variations for a single amf calculation */
    for (is=0; is<output->atm.microphys.nsza_denstab; is++)
      if (fabs (output->atm.microphys.sza_denstab[is]-output->atm.sza_t[iv]) < 1.e-04)
	isthis = is;

    for (lc=0; lc<nlyr; lc++){ 
      output->atm.microphys.dens[output->atm.microphys.denstab_id][lc] = 
	output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc];
    }
  }


  /* loop over wavelength */
  for (iv=output->wl.nlambda_rte_lower; iv<=output->wl.nlambda_rte_upper; iv++) {
    /***********************/
    /* Rayleigh scattering */
    /***********************/

    for (iq=0; iq<output->atm.nq_t[iv]; iq++) {

      /* FU special: The Rayleigh optical depth depends on the subband */
      if (input.ck_scheme == CK_FU) {
#if HAVE_FULIOU
	status = ckdfuray (iv+1, iq+1, cos(output->atm.sza_t[iv]*PI/180.0), nlev,
			   fu_rayleigh);
	
	if (status!=0) {
	  fprintf (stderr, "Error, status %d returned by ckdfuray()\n", status);
	  return status;
	}    
      }
#else
      fprintf (stderr, "Error, Fu and Liou not supported!\n");
      return -1;
#endif  
	
	
      /* loop over layers */
      for (lc=0; lc<nlyr; lc++) {
	
	/* The factor 1.e+5 converts z from km to cm */
	deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;
	
	switch (output->atm.rayleigh) {  /* if Rayleigh scattering is not switched off */
	case RAYLEIGH_CALC:
	  if (input.ck_scheme != CK_FU){
	    bscar = output->atm.cdento[lc] * output->crs.crs[0][MOL_AIR][iv];
	    //fprintf (stderr, "Rayleigh: n dz = %11.7e, crs = %11.7e\n",output->atm.cdento[lc],output->crs.crs[0][MOL_AIR][iv]);
	  }
	  else 
	    bscar = fu_rayleigh[lc];
	  break;
	  
	case RAYLEIGH_FILE_MONO:
	  bscar = output->atm.optprop.tau_rayleigh_user[lc];
	  break;
	  
	case RAYLEIGH_FILE_SPEC:
	  bscar = output->atm.optprop.tau_rayleigh_user_t[iv][lc];
	  break;
	  
	case RAYLEIGH_NONE:
	  bscar = 0;
	  break;
	  
	default:
	  fprintf (stderr, "Error, unknown Rayleigh option %d\n", output->atm.rayleigh);
	  return -1;
	}
	
	output->atm.optprop.tau_rayleigh_t[lc][iv][iq] = bscar;
      }
    }    


    /***********************/
    /* absorption by gases */
    /***********************/
    switch(input.ck_scheme) {
    case CK_NONE:

      iq=0; /* no loop over quadrature points required */
      
      bscartot      = 0;
      babsotot      = 0; 
      babsotot_o3   = 0; 
      babsotot_co2  = 0; 
      babsotot_o2   = 0; 
      babsotot_no2  = 0; 
      babsotot_bro  = 0; 
      babsotot_oclo = 0; 
      babsotot_hcho = 0; 
      babso_amftot  = 0;
      
      vertical_column=0;
  
      // conversion from density to mixing gas ratio
      for ( gas_nr = 1; gas_nr < MOL_NN; gas_nr++) {
  
        if (((input.atm.interpol_method_gas[gas_nr] == INTERP_METHOD_LINMIX) ||
             (input.atm.interpol_method_gas[gas_nr] == INTERP_METHOD_SPLINE) ||        /* spline integration not implemented jet */
             (input.atm.interpol_method_gas[gas_nr] == INTERP_METHOD_LOG_SPLINE)) &&   /* spline integration not implemented jet */
             (output->atm.molabs == MOLABS_CALC) )  { 

          for (lc=0; lc<=nlyr; lc++) 
            output->atm.microphys.dens[gas_nr][lc] /= output->atm.microphys.dens[MOL_AIR][lc];
        
          if (output->atm.microphys.denstab_id == gas_nr) {
	    /* This is airmass factor stuff */
            /* Use column for matrix profile corresponding to the SZA this RTE calculation is done for */
            for (lc=0; lc<=nlyr; lc++) 
	      output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc] /= output->atm.microphys.dens[MOL_AIR][lc];
          }        
        }
      }

      /* loop over layers */
      for (lc=0; lc<nlyr; lc++) {
	
	/* The factor 1.e+5 converts z from km to cm.*/
	deltaz= (output->atm.zd[lc] - output->atm.zd[lc+1])*1e5;
	
	babso      = 0;
	babso_o3   = 0;
	babso_co2  = 0;
	babso_o2   = 0;
	babso_no2  = 0;
	babso_bro  = 0;
	babso_oclo = 0;
	babso_hcho = 0;
	babso_amf  = 0;
	babso_md   = 0;
	
	switch (output->atm.molabs) {
	case MOLABS_CALC: 
          switch (input.atm.z_interpolate) {       
     /*        /\* traditional way (before 2005-11): linear interpolation of the absorption coefficient k_abs = a + bz *\/  */

/* 	    /\* O3 absorption *\/ */
/* 	    babso_o3 += 0.5*deltaz * (output->atm.microphys.dens[MOL_O3][lc]*output->crs.crs[lc][MOL_O3][iv] */
/* 				    + output->atm.microphys.dens[MOL_O3][lc+1]*output->crs.crs[lc+1][MOL_O3][iv]); */
/* 	    babso += babso_o3; */
/* 	    if (output->atm.microphys.denstab_id != MOL_O3) babso_md += babso_o3; */
	  
/* 	    /\* CO2 absorption *\/ */
/* 	    babso_co2 += 0.5*deltaz * (output->atm.microphys.dens[MOL_CO2][lc]*output->crs.crs[lc][MOL_CO2][iv] */
/*                                      + output->atm.microphys.dens[MOL_CO2][lc+1]*output->crs.crs[lc+1][MOL_CO2][iv]); */
/* 	    babso += babso_co2; */
/* 	    if (output->atm.microphys.denstab_id != MOL_CO2) babso_md += babso_co2; */
	  
/* 	    /\* O2 absorption *\/ */
/* 	    babso_o2 += 0.5*deltaz * (output->atm.microphys.dens[MOL_O2][lc]*output->crs.crs[lc][MOL_O2][iv] */
/* 				    + output->atm.microphys.dens[MOL_O2][lc+1]*output->crs.crs[lc+1][MOL_O2][iv]); */
/* 	    babso += babso_o2; */
/* 	    if (output->atm.microphys.denstab_id != MOL_O2) babso_md += babso_o2; */
	  
/* 	    /\* NO2 absorption *\/ */
/* 	    babso_no2 += 0.5*deltaz * (output->atm.microphys.dens[MOL_NO2][lc]*output->crs.crs[lc][MOL_NO2][iv] */
/* 				     + output->atm.microphys.dens[MOL_NO2][lc+1]*output->crs.crs[lc+1][MOL_NO2][iv]); */
/* 	    babso += babso_no2; */
/* 	    if (output->atm.microphys.denstab_id != MOL_NO2) babso_md += babso_no2; */
	  
/* 	    /\* BRO absorption *\/ */
/* 	    babso_bro += 0.5*deltaz * (output->atm.microphys.dens[MOL_BRO][lc]*output->crs.crs[lc][MOL_BRO][iv] */
/* 				     + output->atm.microphys.dens[MOL_BRO][lc+1]*output->crs.crs[lc+1][MOL_BRO][iv]); */
/* 	    babso += babso_bro; */
/* 	    if (output->atm.microphys.denstab_id != MOL_BRO) babso_md += babso_bro; */
	  
/* 	    /\* OCLO absorption *\/ */
/* 	    babso_oclo += 0.5*deltaz * (output->atm.microphys.dens[MOL_OCLO][lc]*output->crs.crs[lc][MOL_OCLO][iv] */
/* 				      + output->atm.microphys.dens[MOL_OCLO][lc+1]*output->crs.crs[lc+1][MOL_OCLO][iv]); */
/* 	    babso += babso_oclo; */
/* 	    if (output->atm.microphys.denstab_id != MOL_OCLO) babso_md += babso_oclo; */
	  
/* 	    /\* HCHO absorption *\/ */
/* 	    babso_hcho += 0.5*deltaz * (output->atm.microphys.dens[MOL_HCHO][lc]*output->crs.crs[lc][MOL_HCHO][iv] */
/* 				      + output->atm.microphys.dens[MOL_HCHO][lc+1]*output->crs.crs[lc+1][MOL_HCHO][iv]); */
/* 	    babso += babso_hcho; */
/* 	    if (output->atm.microphys.denstab_id != MOL_HCHO)  babso_md += babso_hcho; */
	  
/* 	    if (output->atm.microphys.denstab_id) { */
/* 	      /\* Use column for matrix profile corresponding to the SZA this RTE calculation is done for *\/ */
/* 	      /\* This is airmass factor stuff *\/ */
/* 	      babso_amf += 0.5*deltaz * (output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc]* */
/*                                          output->crs.crs[lc][output->atm.microphys.denstab_id][iv] */
/* 				       + output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc+1]* */
/* 				         output->crs.crs[lc+1][output->atm.microphys.denstab_id][iv]); */
/* 	      vertical_column += 0.5*deltaz * (output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc] */
/* 					     + output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc+1]); */
/* 	      crs_amf = output->crs.crs[lc+1][output->atm.microphys.denstab_id][iv]; */
/* 	    } */
/* 	    break; */
          case Z_INTERPOLATE_LOGLIN:
          case Z_INTERPOLATE_LINMIX:
          case Z_INTERPOLATE_SPLINE:

            for (gas_nr = 1; gas_nr < MOL_NN; gas_nr++) {

              switch (input.atm.interpol_method_gas[gas_nr]) {
                case INTERP_METHOD_LINEAR:
		  density = 0.5 * (output->atm.microphys.dens[gas_nr][lc] + output->atm.microphys.dens[gas_nr][lc+1]);
                  break;
	        case INTERP_METHOD_LOG:
                  density = log_average(output->atm.microphys.dens[gas_nr][lc],output->atm.microphys.dens[gas_nr][lc+1]);
                  break;
                case INTERP_METHOD_LINMIX:
	        case INTERP_METHOD_SPLINE:
                case INTERP_METHOD_LOG_SPLINE:
                  density = linmix_average(output->atm.microphys.dens[gas_nr][lc] ,output->atm.microphys.dens[gas_nr][lc+1],
                                           output->atm.microphys.dens[MOL_AIR][lc],output->atm.microphys.dens[MOL_AIR][lc+1]);
                  break;
	        default:
	          fprintf (stderr, "Error, unknown z_interpol_method_gas option %d\n",input.atm.interpol_method_gas[gas_nr]);
	          return -1;
	      }

              switch (gas_nr) {
	        case MOL_O3: 
                  babso_o3   += deltaz * 0.5 * (output->crs.crs[lc][MOL_O3][iv]   + output->crs.crs[lc+1][MOL_O3][iv])   * density;
                  babso += babso_o3;
                  if (output->atm.microphys.denstab_id != MOL_O3)   babso_md += babso_o3;
	          break;
                case MOL_CO2: 
                  babso_co2  += deltaz * 0.5 * (output->crs.crs[lc][MOL_CO2][iv]  + output->crs.crs[lc+1][MOL_CO2][iv])  * density;
                  babso += babso_co2;
                  if (output->atm.microphys.denstab_id != MOL_CO2)  babso_md += babso_co2;
 	          break;
	        case MOL_O2:
	          babso_o2   += deltaz * 0.5 * (output->crs.crs[lc][MOL_O2][iv]   + output->crs.crs[lc+1][MOL_O2][iv])   * density;
	          babso += babso_o2;
	          if (output->atm.microphys.denstab_id != MOL_O2)   babso_md += babso_o2;
	          break;
                case MOL_NO2:
                  babso_no2  += deltaz * 0.5 * (output->crs.crs[lc][MOL_NO2][iv]  + output->crs.crs[lc+1][MOL_NO2][iv])  * density;
                  babso += babso_no2;
                  if (output->atm.microphys.denstab_id != MOL_NO2)  babso_md += babso_no2;
	          break;
                case MOL_BRO: 
                  babso_bro  += deltaz * 0.5 * (output->crs.crs[lc][MOL_BRO][iv]  + output->crs.crs[lc+1][MOL_BRO][iv])  * density;
                  babso += babso_bro;
                  if (output->atm.microphys.denstab_id != MOL_BRO)  babso_md += babso_bro; 
	          break;
                case MOL_OCLO:               
	          babso_oclo += deltaz * 0.5 * (output->crs.crs[lc][MOL_OCLO][iv] + output->crs.crs[lc+1][MOL_OCLO][iv]) * density;
	          babso += babso_oclo;
	          if (output->atm.microphys.denstab_id != MOL_OCLO) babso_md += babso_oclo;
	          break;
	        case MOL_HCHO:
	          babso_hcho += deltaz * 0.5 * (output->crs.crs[lc][MOL_HCHO][iv] + output->crs.crs[lc+1][MOL_HCHO][iv]) * density;
	          babso += babso_hcho;
	          if (output->atm.microphys.denstab_id != MOL_HCHO) babso_md += babso_hcho;
	          break;
              }

	      if (output->atm.microphys.denstab_id == gas_nr) {
	        /* This is airmass factor stuff */
	        /* Use column for matrix profile corresponding to the SZA this RTE calculation is done for */
                switch (input.atm.interpol_method_gas[gas_nr]) {
                case INTERP_METHOD_LINEAR:
                  density =         0.5 * (output->atm.microphys.denstab[gas_nr][isthis][lc] + 
                                           output->atm.microphys.denstab[gas_nr][isthis][lc+1]);
		  break;
	        case INTERP_METHOD_LOG:
                  density = log_average   (output->atm.microphys.denstab[gas_nr][isthis][lc],
                                           output->atm.microphys.denstab[gas_nr][isthis][lc+1]);
		  break;
                case INTERP_METHOD_LINMIX:
	        case INTERP_METHOD_SPLINE:
                case INTERP_METHOD_LOG_SPLINE: 
                  density = linmix_average(output->atm.microphys.denstab[gas_nr][isthis][lc],
                                           output->atm.microphys.denstab[gas_nr][isthis][lc+1],
                                           output->atm.microphys.dens[MOL_AIR][lc],
                                           output->atm.microphys.dens[MOL_AIR][lc+1]);
		  break;
                default:
	          fprintf (stderr, "Error: unkown z_interpol_gas option for denstab calculation \n");
	          return -1;
	          break;
                }
              
	        babso_amf += deltaz * 0.5 * (output->crs.crs[lc][gas_nr][iv] +
                                             output->crs.crs[lc+1][gas_nr][iv]) * density;
	        vertical_column += deltaz * density;
	        crs_amf = output->crs.crs[lc+1][output->atm.microphys.denstab_id][iv];
	      }
	    }
            break;
          default:
	    fprintf (stderr, "Error: unkown z_interpolate option %d \n", input.atm.z_interpolate);
	    return -1;
	    break;
	  }

	  output->atm.optprop.tau_molabs_t[lc][iv][iq]    = babso;
	  output->atm.optprop.tau_molabs_md_t[lc][iv][iq] = babso_md;
	  break;
	  
	case MOLABS_FILE_MONO:
	  babso = output->atm.optprop.tau_molabs_user[lc];
	  output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
	  break;
	  
	case MOLABS_FILE_SPEC:
	  babso = output->atm.optprop.tau_molabs_user_t[iv][lc];
	  output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
	  break;
	  
	case MOLABS_NONE:
	  babso = 0;
	  output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
	  break;
	  
	default:
	  fprintf (stderr, "Error, unknown molecular absorption option %d\n", output->atm.molabs);
	  return -1;
	}
	
	output->atm.wght_t[iv][iq] = 1;  /* default weight */
      

	if (input.verbose) {
	  bscar=output->atm.optprop.tau_rayleigh_t[lc][iv][iq];
	  bscartot+=bscar;
	  
	  babsotot      += babso;
	  babsotot_o3   += babso_o3;
	  babsotot_co2  += babso_co2;
	  babsotot_o2   += babso_o2;
	  babsotot_no2  += babso_no2;
	  babsotot_bro  += babso_bro;
	  babsotot_oclo += babso_oclo;
	  babsotot_hcho += babso_hcho;
	  babso_amftot  += babso_amf;
	  
	  if ( lc ==0 ) {
	    fprintf (stderr, "\n*** wavelength: iv = %d, %f nm\n", iv, output->wl.lambda_t[iv]);
	    fprintf (stderr, "*** Rayleigh scattering cross section: %e\n", output->crs.crs[0][MOL_AIR][iv]);
	    fprintf (stderr, "*** setup_gases()\n");
	    fprintf (stderr, " ----------------------------------------------------------------------------------------------------------------------------------------------\n");
	    fprintf (stderr, "   lc |    z[km] |  Rayleigh   |   Molecular \n");
	    fprintf (stderr, "      |          |    dtau     |   absorption        o3            co2           o2           no2            bro           oclo          hcho\n");
	    fprintf (stderr, " ----------------------------------------------------------------------------------------------------------------------------------------------\n");
	  }		  
	  fprintf (stderr, "%5d | %8.2f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f | %11.6f\n", 
		   lc, output->atm.zd[lc+1],
		   bscar,
		   babso, babso_o3, babso_co2, babso_o2, babso_no2, babso_bro, babso_oclo, babso_hcho);
	}
      } /* End layer loop */

      if (input.verbose) {
	fprintf (stderr, " ----------------------------------------------------------------------------------------------------------------------------------------------\n");
	fprintf (stderr, "%5s | %8.2f | %11.6f | %11.6f | %12.6e| %12.6e| %12.6e| %12.6e| %12.6e| %12.6e| %12.6e\n", 
		 "sum", 0.0/0.0,
		 bscartot,
		 babsotot, babsotot_o3, babsotot_co2, babsotot_o2, babsotot_no2, babsotot_bro, babsotot_oclo, babsotot_hcho);
	fprintf (stderr, " ----------------------------------------------------------------------------------------------------------------------------------------------\n");
      }


      // conversion back from mixing gas ratios to density 
      for ( gas_nr = 1; gas_nr < MOL_NN; gas_nr++) {
  
        if (((input.atm.interpol_method_gas[gas_nr] == INTERP_METHOD_LINMIX) ||
             (input.atm.interpol_method_gas[gas_nr] == INTERP_METHOD_SPLINE) ||       /* spline integration not implemented jet */
             (input.atm.interpol_method_gas[gas_nr] == INTERP_METHOD_LOG_SPLINE)) &&   /* spline integration not implemented jet */
             (output->atm.molabs == MOLABS_CALC) )  {  

          for (lc=0; lc<=nlyr; lc++) 
            output->atm.microphys.dens[gas_nr][lc] *= output->atm.microphys.dens[MOL_AIR][lc];
        
          if (output->atm.microphys.denstab_id == gas_nr) {
            /* Use column for matrix profile corresponding to the SZA this RTE calculation is done for */
	    /* This is airmass factor stuff */
            for (lc=0; lc<=nlyr; lc++) 
	      output->atm.microphys.denstab[output->atm.microphys.denstab_id][isthis][lc] *= output->atm.microphys.dens[MOL_AIR][lc];
          }        
        }
      }

      break;

    case CK_KATO:
    case CK_KATO2:
    case CK_KATO2_96:
      nq1 = output->crs_ck.profile[KATO_CO2-1][iv].ngauss;
      nq2 = output->crs_ck.profile[KATO_O3-1 ][iv].ngauss;
      nq3 = output->crs_ck.profile[KATO_O2-1 ][iv].ngauss;
      nq4 = output->crs_ck.profile[KATO_H2O-1][iv].ngauss;
      
      if (input.verbose) {
	fprintf (stderr, "*** subbands:\n");
	fprintf (stderr, "***  CO2: %2d\n", nq1);
	fprintf (stderr, "***  O3:  %2d\n", nq2);
	fprintf (stderr, "***  O2:  %2d\n", nq3);
	fprintf (stderr, "***  H2O: %2d\n", nq4);
      }
      
      for (lc=0; lc<output->atm.nlyr; lc++) {
	
	/* The factor 1e5 converts z from km to cm.*/
	deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;
      
	iq=0;  /* reset quadrature counter */
	/* loop over components (4D correlated-k) */
	for (iq1=0; iq1<nq1; iq1++) {  /* CO2 */
	  
	  /* set CO2 cross section and weights */
	  output->crs.crs[lc][MOL_CO2][iv] = 
	    output->crs_ck.profile[KATO_CO2-1][iv].crs[lc][iq1];
	  
	  
	  weight[KATO_CO2-1] = output->crs_ck.profile[KATO_CO2-1][iv].weight[iq1];
	  
	  for (iq2=0; iq2<nq2; iq2++) { /* O3 */
	    
	    /* set O3 cross section and weights */
	    output->crs.crs[lc][MOL_O3][iv]  = 
	      output->crs_ck.profile[KATO_O3-1][iv].crs[lc][iq2];
	    
	    weight[KATO_O3-1] = output->crs_ck.profile[KATO_O3-1][iv].weight[iq2];
	    
	    for (iq3=0; iq3<nq3; iq3++) { /* O2 */
	      
	      /* set O2 cross section and weights */
	      output->crs.crs[lc][MOL_O2][iv]  =
		output->crs_ck.profile[KATO_O2-1][iv].crs[lc][iq3];
	      
	      weight[KATO_O2-1] = output->crs_ck.profile[MOL_O2][iv].weight[iq3];
	      
	      for (iq4=0; iq4<nq4; iq4++) { /* H2O */
		
		/* set H2O cross section and weights */
		output->crs.crs[lc][MOL_H2O][iv]  = 
		  output->crs_ck.profile[KATO_H2O-1][iv].crs[lc][iq4];
		
		weight[KATO_H2O-1] = output->crs_ck.profile[MOL_H2O][iv].weight[iq4];
		
		/* quadrature weight */
		factor = weight[KATO_CO2-1]*weight[KATO_O3-1]*weight[KATO_O2-1]*weight[KATO_H2O-1];
		

		/* sum absorption */
		babso = 0;
		
		/* CO2 absorption */
		babso += deltaz * output->crs.crs[lc][MOL_CO2][iv] * 
		  0.5 * (output->atm.microphys.dens[MOL_CO2][lc] + output->atm.microphys.dens[MOL_CO2][lc+1]);
		
		/* O3 absorption */
		babso += deltaz * output->crs.crs[lc][MOL_O3][iv]  *
		  0.5 * (output->atm.microphys.dens[MOL_O3][lc]  + output->atm.microphys.dens[MOL_O3][lc+1]);
		
		/* O2 absorption */
		babso += deltaz * output->crs.crs[lc][MOL_O2][iv]  *
		  0.5 * (output->atm.microphys.dens[MOL_O2][lc]  + output->atm.microphys.dens[MOL_O2][lc+1]);
		
		/* H2O absorption */
		babso += deltaz * output->crs.crs[lc][MOL_H2O][iv] *
		  0.5 * (output->atm.microphys.dens[MOL_H2O][lc] + output->atm.microphys.dens[MOL_H2O][lc+1]);
		
		/* copy absorption cross section and weight */
		/* to final destination                     */
		output->atm.wght_t[iv][iq] = factor;
		
		output->atm.optprop.tau_molabs_t[lc][iv][iq]   = babso;
		iq++;
		
	      }  /* end of correlated-k (iq4) loop */
	    }  /* end of correlated-k (iq3) loop */
	  }  /* end of correlated-k (iq2) loop */
	}  /* end of correlated-k (iq1) loop */
      }
  
      break;
	
    case CK_FU:
    case CK_AVHRR_KRATZ:
    case CK_GENERIC:
    case CK_LOWTRAN:

      nq = output->crs_ck.profile[0][iv].ngauss;

      /* loop over quadrature points */
      
      for (iq=0; iq<nq; iq++) {
	
	/* set absorption coefficient */
	for (lc=0; lc<output->atm.nlyr; lc++) {

	  /* The factor 1e5 converts z from km to cm.*/
	  deltaz = (output->atm.zd[lc] - output->atm.zd[lc+1]) * 1e5;

	  output->atm.optprop.tau_molabs_t[lc][iv][iq] =
	    deltaz/1e5 * output->crs_ck.profile[0][iv].crs[lc][iq];
	}
	  
	/* quadrature weight */
	factor = output->crs_ck.profile[0][iv].weight[iq];
	output->atm.wght_t[iv][iq] = factor;
      }
      
      break;
      
    default:
      fprintf (stderr, "Error: unimplemented correlated-k scheme\n");
      return -1;
      break;
    }
  } /* End wavelength loop */

  free(fu_rayleigh);


  return 0;  /* ok */
}



/**************************************************************/
/* Setup temperature at output levels.                        */
/**************************************************************/

int setup_temperature (input_struct input, output_struct *output)
{
  int status=0, lev=0, iz=0;
  float *tmp_temper=NULL, *tmp_zd=NULL, *zout=NULL;

  /* Linearly interpolate output temperature to zout */
  /* sort profile in ascending order, assuming that */
  /* the original profile was descending            */
  
  tmp_zd     = (float *) calloc (output->atm.nlev, sizeof(float));
  tmp_temper = (float *) calloc (output->atm.nlev, sizeof(float));
  
  /* sort temperature profile in ascending order */
  for (lev=0; lev<output->atm.nlev; lev++) {
    tmp_zd[lev]     = output->atm.zd     [output->atm.nlev-1-lev];
    tmp_temper[lev] = output->atm.microphys.temper [output->atm.nlev-1-lev];
  }

  output->atm.temper_out  = (float *) calloc (input.atm.nzout, sizeof(float));
 
  /* Now we need to define a temporary zout grid;        */
  /* for rte_solver montecarlo it is possible to         */
  /* define an output altitude of -999 which stands      */
  /* for the surface altitude which may be a function    */
  /* of (x,y). This value is replaced by 0 which means   */
  /* that the surface temperature at the surface output  */
  /* level is set to temperature(z=0) which is not quite */
  /* correct if a 2D altitude grid has been defined.     */
  /* However, the only place where this information is   */
  /* currently used is the flexstor header which is      */
  /* required for photolysis calculations.               */

  zout = calloc (output->atm.nzout, sizeof(float));
  for (iz=0; iz<output->atm.nzout; iz++) {
    if (output->atm.zout_sea[iz] != -999)
      zout[iz] = output->atm.zout_sea[iz];
    else
      zout[iz] = 0;
  }
  
  /* interpolate temperature to zout grid */
  status = arb_wvn (output->atm.nlev, tmp_zd, tmp_temper, 
		    input.atm.nzout, zout, output->atm.temper_out, 1, 0);

  if (status != 0) {
    fprintf (stderr, "Error interpolating temperature to output grid\n");
    return status;
  }


  free(tmp_temper); 
  free(tmp_zd);
  free(zout);


  /* Finally, copy the temperature of the lowest level to */
  /* output->temperature if the latter is not already set */
  /* ATTENTION: moved this code to redistribute.c because */
  /* we want the surface temperature to equal the         */
  /* z=altitude rather than z=0                           */

  /*
  if (output->surface_temperature < 0)
    output->surface_temperature = output->atm.microphys.temper[output->atm.nlev-1];
  */

  return 0;
}




/**************************************************************/
/* Read atmosphere file and write to atm_out_struct           */
/**************************************************************/

static int read_atmosphere (char *filename, atm_out_struct *out, 
                            int interpol_method_press, int interpol_method_temper, int *interpol_method_gas,
                            int cdf, int quiet, int verbose) 

{
  int i=0, rows=0, min_columns=0, max_columns=0;
  int lev=0, status = 0;
  float **atm_data=NULL;
  double boltzmann = 1.38065e-23;

  size_t dimlen=0;
  
  int nlev=0, nlyr=0;

  float tmpdens=0, idealdiff=0;
  int idealgas=0, firstnonideal=0;


#if HAVE_LIBNETCDF
  int ncid=0;
  int idd_nlev=0;
  int id_z=0, id_press=0, id_temper=0;
  int id_air=0, id_o3=0, id_o2=0, id_h2o=0, id_co2=0;
  
  size_t start[1] = {0};
  size_t count[1] = {0};
#endif

  if (cdf==0) {   /* classic ASCII atmospheric model */
    status = ASCII_file2float (filename, 
			       &rows, &max_columns, &min_columns, 
			       &atm_data);

    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    if (max_columns!=min_columns) {
      fprintf (stderr, "Warning: Inconsistent number of columns in %s\n", filename);
      fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
      return -1;
    }  
    
    if (min_columns<3) {
      fprintf (stderr, "Error, too few columns in %s! Need at least %d.\n", filename, 3);
      return -1;
    }
        
    nlev = rows;
    out->nlev = nlev;
    nlyr = rows-1;
  
    /* allocate memory for density array */
    out->microphys.dens = calloc(MOL_NN, sizeof(float *));
    if (out->microphys.dens == NULL) {
      fprintf (stderr, "Error allocating memory for dens\n");
      return -1;
    }
    
    out->microphys.press  = ASCII_column_float (atm_data, nlev, ATMFILE_PRESS );
    out->microphys.temper = ASCII_column_float (atm_data, nlev, ATMFILE_TEMPER );
    out->zd               = ASCII_column_float (atm_data, nlev, ATMFILE_ZD );


    /**************************************/
    /* the following columns are optional */
    /**************************************/

    if (min_columns>3) { /* air */
      out->microphys.dens[MOL_AIR] = ASCII_column_float(atm_data, nlev, ATMFILE_AIR );
    }
    else {
      out->microphys.dens[MOL_AIR] = (float *) calloc (nlev, sizeof(float));
      
      /* Set air density */
      for (lev=0; lev<nlev; lev++)
	out->microphys.dens[MOL_AIR][lev] = out->microphys.press[lev]*1E-04/(boltzmann*out->microphys.temper[lev]);
    }
    
    if (min_columns>4) /* o3   */
      out->microphys.dens[MOL_O3]  = ASCII_column_float(atm_data, nlev, ATMFILE_O3 );
    else
      out->microphys.dens[MOL_O3]  = (float *) calloc (nlev, sizeof(float));
    
    if (min_columns>5)  /* oxygen           */
      out->microphys.dens[MOL_O2]  = ASCII_column_float(atm_data, nlev, ATMFILE_O2 );
    else 
      out->microphys.dens[MOL_O2]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>6)  /* water vapour     */
      out->microphys.dens[MOL_H2O] = ASCII_column_float(atm_data, nlev, ATMFILE_H2O);
    else 
      out->microphys.dens[MOL_H2O]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>7)  /* carbon dioxide   */
      out->microphys.dens[MOL_CO2] = ASCII_column_float(atm_data, nlev, ATMFILE_CO2);
    else 
      out->microphys.dens[MOL_CO2]  = (float *) calloc (nlev, sizeof(float));
    
    
    if (min_columns>8)  /* nitrogen dioxide */
      out->microphys.dens[MOL_NO2] = ASCII_column_float(atm_data, nlev, ATMFILE_NO2);
    else 
      out->microphys.dens[MOL_NO2]  = (float *) calloc (nlev, sizeof(float));

    /* Allocate memory for species not in atmosphere file */
    out->microphys.dens[MOL_BRO]  = (float *) calloc (nlev, sizeof(float));
    out->microphys.dens[MOL_OCLO] = (float *) calloc (nlev, sizeof(float));
    out->microphys.dens[MOL_HCHO] = (float *) calloc (nlev, sizeof(float));

    ASCII_free_float(atm_data, rows);
  }
  else {

#if HAVE_LIBNETCDF

    /* open netcdf file */
    status = nc_open (filename, NC_NOWRITE, &ncid);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
      return status;
    }

    /* get dimension id for "nlev" */
    status = nc_inq_dimid (ncid, "nlev", &idd_nlev);
  
    /* get dimension length for "nlev" */
    status = nc_inq_dimlen (ncid, idd_nlev, &dimlen);
    nlev = dimlen;
    nlyr = dimlen-1;


    start[0] = 0;
    count[0] = nlev;

    /* allocate memory for profiles */
    out->zd               = calloc (nlev, sizeof (float));
    out->microphys.press  = calloc (nlev, sizeof (float));
    out->microphys.temper = calloc (nlev, sizeof (float));


    /* get variable id for "z" */
    status = nc_inq_varid (ncid, "z", &id_z);
    
    /* read altitude "z" */
    status = nc_get_vara_float (ncid, id_z, start, count, out->zd);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading z from %s\n", status, filename);
      return status;
    }


    /* get variable id for "pressure" */
    status = nc_inq_varid (ncid, "pressure", &id_press);
    
    /* read "pressure" */
    status = nc_get_vara_float (ncid, id_press, start, count, out->microphys.press);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading pressure from %s\n", status, filename);
      return status;
    }


    /* get variable id for "temperature" */
    status = nc_inq_varid (ncid, "temperature", &id_temper);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading temperature from %s\n", status, filename);
      return status;
    }
    
    /* read "temperature" */
    status = nc_get_vara_float (ncid, id_temper, start, count, out->microphys.temper);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading temperature from %s\n", status, filename);
      return status;
    }


    /* allocate memory for density array */
    out->microphys.dens = calloc(MOL_NN, sizeof(float *));
    if (out->microphys.dens == NULL) {
      fprintf (stderr, "Error allocating memory for dens\n");
      return -1;
    }

    for (i=0; i<MOL_NN; i++) {
      out->microphys.dens[i] = calloc(nlev, sizeof(float));
      if (out->microphys.dens[i] == NULL) {
	fprintf (stderr, "Error allocating memory for dens[%d]\n", i);
	return -1;
      }
    }


    /* get variable id for "density" */
    status = nc_inq_varid (ncid, "density", &id_air);
    
    /* read "density" */
    status = nc_get_vara_float (ncid, id_air, start, count, out->microphys.dens[MOL_AIR]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading density from %s\n", status, filename);
      return status;
    }


    /* get variable id for "o3" */
    status = nc_inq_varid (ncid, "o3", &id_o3);
    
    /* read "o3" */
    status = nc_get_vara_float (ncid, id_o3, start, count, out->microphys.dens[MOL_O3]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading o3 from %s\n", status, filename);
      return status;
    }


    /* get variable id for "o2" */
    status = nc_inq_varid (ncid, "o2", &id_o2);
    
    /* read "o2" */
    status = nc_get_vara_float (ncid, id_o2, start, count, out->microphys.dens[MOL_O2]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading o2 from %s\n", status, filename);
      return status;
    }


    /* get variable id for "h2o" */
    status = nc_inq_varid (ncid, "h2o", &id_h2o);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading h2o from %s\n", status, filename);
      return status;
    }
    
    /* read "h2o" */
    status = nc_get_vara_float (ncid, id_h2o, start, count, out->microphys.dens[MOL_H2O]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading h2o from %s\n", status, filename);
      return status;
    }


    /* get variable id for "co2" */
    status = nc_inq_varid (ncid, "co2", &id_co2);
    
    /* read "co2" */
    status = nc_get_vara_float (ncid, id_co2, start, count, out->microphys.dens[MOL_CO2]);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading co2 from %s\n", status, filename);
      return status;
    }


    /* we don't have no2 in the netcdf file, hence we keep it 0;        */
    /* also we don't read n2o, co, and ch4 because we never used        */
    /* those before (should change that by introducing all relevant     */
    /* gases into dens; should also get rid of ndens  */
    /* and replace that by MOL_NN)                                      */

    nc_close (ncid);
    
#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot            *\n");
  fprintf (stderr, " * use the generic correlated_k option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif

  }


  /* Finally, check if profiles of pressure, temperature and air density are consistent */
  /* with ideal gas law over 6 orders of magnitude in pressure; check only first six    */
  /* orders of magnitude because for very small pressures roundoff errors due to too    */
  /* few digits cause problems                                                          */
  firstnonideal=1;
  for (lev=0; lev<nlev; lev++) {

    idealgas=1;

    if (out->microphys.press[lev]/out->microphys.press[nlev-1]>1e-6) {
      tmpdens = out->microphys.press[lev]*1E-04 / (boltzmann*out->microphys.temper[lev]);
      
      if ((tmpdens==0 && out->microphys.dens[MOL_AIR][lev]!=0) || 
	  (tmpdens!=0 && out->microphys.dens[MOL_AIR][lev]==0))
	idealgas=0;
      
      if (out->microphys.dens[MOL_AIR][lev]>0)
	if ((idealdiff = fabs ((tmpdens - out->microphys.dens[MOL_AIR][lev]) / out->microphys.dens[MOL_AIR][lev])) > 0.01)
	  idealgas=0;
      
      if (!idealgas && firstnonideal) {
	firstnonideal=0;
	fprintf (stderr, "\n");
	fprintf (stderr, "*** ATTENTION: pressure, temperature, and air density not\n");
	fprintf (stderr, "*** consistent with ideal gas law at the following levels\n");
	fprintf (stderr, "*** (assuming a Boltzmann constant %.6e):\n", boltzmann);
	fprintf (stderr, "***    z [km]  density [cm-3]  p/kT [cm-3]  difference\n");
      }
      
      if (!idealgas) {
	fprintf (stderr, "***   %7.2f  %.4e      %.4e    %6.1f%%\n", 
		 out->zd[lev], out->microphys.dens[MOL_AIR][lev], tmpdens, idealdiff*100.0);
      }
    }
  }
  
  if (!firstnonideal) {
    fprintf (stderr, "*** Please check pressure, temperature, and air density in\n");
    fprintf (stderr, "*** %s for consistency!\n", filename);
    fprintf (stderr, "\n");
    return -1;
  } 

  return 0;
}


/**************************************************************/
/* Read density file                                          */
/* Combine with background atmosphere and                     */
/* Interpolate on the common grid: z_all                      */
/**************************************************************/

static int read_and_combine_density (char *filename, atm_out_struct *out,
				     int dens_id, int *interpol_method_gas,
				     float *dens_air_all, float *z_all, int nlev_all,
                                     float *dens_air_atm, float *z_atm, int nlev_atm,
                                     int solver, int quiet, int verbose)

{
  int rows=0, min_columns=0, max_columns=0, nlev=0;
  int status=0;
  float **atm_data=NULL;
  int lc=0;
  int i; 
  float *zd=NULL;
  float **dens=NULL;
  static int first_dens = 1;
  float *z_comb=NULL;
  float **dens_comb=NULL;
  int nlev_comb=-666;
  float *dens_air_comb=NULL;
  int iv=0, is=0, isthis=0;
  
  if ((strlen(filename) == 0) && (out->nlev == nlev_all)) {
    /*nothing to do, everything is fine*/
    return 0;
  }
  else { /* something to do */
    
    /* densfile specified in the input file */
    if (strlen(filename) > 0){

      if (verbose)
        fprintf (stderr," ... reading density file %s\n", filename);

      /* read densfile and combine it with background density*/

      status = ASCII_file2float (filename, 
                                 &rows, &max_columns, &min_columns, 
                                 &atm_data);
  
      if (status!=0) {
        fprintf (stderr, "Error %d reading %s\n", status, filename);
        return status;
      }

      if (max_columns==2) {   /* 2-columns: 1 altitude, 2 density */
    
        /* reading data from file */
        if (max_columns!=min_columns ) {
          fprintf (stderr, "Error: Inconsistent number of columns in %s\n", filename);
          fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
          fprintf (stderr, "Should have been %d. First altitude, second density.\n", 2);
          return -1;
        }
 
        nlev  = rows;
        zd   = ASCII_column_float (atm_data, nlev, 0 );

        status  = ASCII_calloc_float(&dens, 1, nlev);
        if (status!=0) {
	  fprintf (stderr, "Error %d allocating memory for dens\n", status); 
	  return status; 
        }

        for (lc=0; lc<rows; lc++)
          dens[0][lc] = atm_data[lc][1];
       
        /* combine file data and background */
        combine_profiles (dens, zd, nlev, 1,
                          out->microphys.dens[dens_id], out->zd, out->nlev,
                          interpol_method_gas[dens_id],
                          &(dens_comb), &(z_comb), &(nlev_comb));

        /* check, if dens = 0 in the uppermost level, for log-interpolation types */
        if ( interpol_method_gas[dens_id] == INTERP_METHOD_LOG || 
             interpol_method_gas[dens_id] == INTERP_METHOD_LOG_SPLINE ) {
          for (i=0; i<1; i++)
            if (dens_comb[i][0] == 0.0 ) {
              fprintf (stderr, "Error: dens(%5.1f km) = 0.0 cannot be interpolated with log assumption\n",z_comb[0]);
              fprintf (stderr, "  ***  By removing this line from the file %s, values for\n", filename);
              fprintf (stderr, "  ***  hights above the last entry in densfile will be \n");
              fprintf (stderr, "  ***  filled with the values given by the background atmosphere.\n");
              return -1;
            }
        }
        
        /* first determine air_dens profile on z_comb for linmix interpolation */
        if (interpol_method_gas[dens_id] == INTERP_METHOD_LINMIX)
          status += arb_wvn (nlev_all, z_all,  dens_air_all, 
                             nlev_comb,  z_comb, dens_air_comb, 
                             interpol_method_gas[MOL_AIR], 1);

        /* interpolate combined data on the z_all grid */
        status += interpolate_density (z_comb, &(dens_comb[0]), nlev_comb, 
                                       z_all, nlev_all, interpol_method_gas[dens_id],
                                       dens_air_comb, dens_air_all, quiet);

        if (status!=0) {
          fprintf (stderr, "Error %d interpolating density to common atmospheric grid\n", status);
          return status;
        }

        /* put result on final array */
        free(out->microphys.dens[dens_id]);
        out->microphys.dens[dens_id] = (float *) calloc (nlev_all, sizeof (float));
    
        for (lc=0; lc<nlev_all; lc++)
          out->microphys.dens[dens_id][lc] = dens_comb[0][lc];
      }

      else if (max_columns > 2 ) {   /* density matrix for airmass calculations */

        if (max_columns!=min_columns) {
          fprintf (stderr, "Error: Inconsistent number of columns in %s\n", filename);
          fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
          return -1;
        }

        if (solver != SDISORT) {
          fprintf (stderr, "Error: Dens-file %s has the format of a matrix.\n", filename);
          fprintf (stderr, " ***   This input type is only possible for solver SDISORT.\n");
          return -1;
        }

    
        /* first row/column contains sza/altitude */
        out->microphys.nsza_denstab = max_columns-1;
        nlev = rows-1;

        zd  = (float *) calloc (nlev, sizeof (float));

        if (out->microphys.denstab_id > 0 ) {
          fprintf (stderr, "Error: dens_file with matrix instead of profile\n");
          fprintf (stderr, "       already specified. The matrix may be used\n");
          fprintf (stderr, "       only for one specie. Profiles for as many as you like\n");
          return -1;
        }

        if (first_dens) {
          out->microphys.denstab_id  = dens_id;
          out->microphys.sza_denstab = (float *) calloc (out->microphys.nsza_denstab, sizeof (float));

          status  = ASCII_calloc_float(&out->microphys.denstab_amf,
                                       out->microphys.nsza_denstab, nlev);
          if (status!=0) {
            fprintf (stderr, "Error %d allocating memory for out->denstab_amf\n", status);
            return status;
          }
        }

        /* copy first row to solar zenith angle array  */
        for (i=1; i<=out->microphys.nsza_denstab; i++)
          out->microphys.sza_denstab[i-1] = atm_data[0][i];
    
        /* copy first column to altitude array zd */
        for (lc=1; lc<=nlev; lc++)
          zd[lc-1] = atm_data[lc][0];
  
        /* copy the remainder to 2-dim double array denstab_amf */
        for (i=1; i<=out->microphys.nsza_denstab; i++)
          for (lc=1; lc<=nlev; lc++)
            out->microphys.denstab_amf          [i-1][lc-1] = atm_data[lc][i];

        /* combine file data and background */
        combine_profiles (out->microphys.denstab_amf, zd, nlev, out->microphys.nsza_denstab,
                          out->microphys.dens[dens_id], out->zd, out->nlev,
                          interpol_method_gas[dens_id],
                          &(dens_comb), &(z_comb), &(nlev_comb));

/* 	for (lc=1; lc<=nlev_all; lc++) */
/*           fprintf (stderr, "lc = %d, z_comb = %5.2f, dens_comb %11.7e\n", lc, z_comb[lc], dens_comb[0][lc]); */

        /* check, if dens = 0 in the uppermost level, for log-interpolation types */
        if ( interpol_method_gas[dens_id] == INTERP_METHOD_LOG || 
             interpol_method_gas[dens_id] == INTERP_METHOD_LOG_SPLINE ) {
          for (i=0; i<out->microphys.nsza_denstab; i++)
            if (dens_comb[i][0] == 0.0 ) {
              fprintf (stderr, "Error: dens(%5.1f km) = 0.0 cannot be interpolated with log assumption\n",z_comb[0]);
              fprintf (stderr, "  ***  By removing this line from the file %s, values for\n", filename);
              fprintf (stderr, "  ***  hights above the last entry in densfile will be \n");
              fprintf (stderr, "  ***  filled with the values given by the background atmosphere.\n");
              return -1;
            }
        }

        /* first determine air_dens profile on z_comb for linmix interpolation */
        if (interpol_method_gas[dens_id] == INTERP_METHOD_LINMIX)
          status += arb_wvn (nlev_all, z_all,  dens_air_all, 
                             nlev_comb,  z_comb, dens_air_comb, 
                             interpol_method_gas[MOL_AIR], 1);


        /* interpolate combined data on the z_all grid */
        for (i=0; i<out->microphys.nsza_denstab; i++){
          status += interpolate_density (z_comb, &(dens_comb[i]), nlev_comb, 
                                         z_all, nlev_all, interpol_method_gas[dens_id],
                                         dens_air_comb, dens_air_all, quiet);

          if (status!=0) {
            fprintf (stderr, "Error %d interpolating denstab (dens_comb) to common atmospheric grid\n", status); 
            return status;
          }
        }

        /* reallocate denstab_amf*/
        ASCII_free_float(out->microphys.denstab_amf, out->microphys.nsza_denstab);
        status  = ASCII_calloc_float(&out->microphys.denstab_amf,
                                      out->microphys.nsza_denstab, nlev_all);
        if (status!=0) {
          fprintf (stderr, "Error %d allocating memory for out->denstab_amf\n", status);
          return status;
        }

        /* copy combined data on final place  */
        for (i=0; i<out->microphys.nsza_denstab; i++)
          for (lc=0;lc<nlev_all;lc++)
            out->microphys.denstab_amf[i][lc]=dens_comb[i][lc];
           
        /* callocate denstab array */
        if (first_dens) {
          status  = ASCII_calloc_float_3D (&out->microphys.denstab, MOL_NN, out->microphys.nsza_denstab, nlev_all);
          if (status!=0) {
            fprintf (stderr, "Error %d allocating memory for out->denstab\n", status);
            return status;
          }
        }

        /* copy the combined 2-dim double array to denstab */
        for (i=0; i<out->microphys.nsza_denstab; i++)
          for (lc=0; lc<nlev_all; lc++)
            out->microphys.denstab [dens_id][i][lc] = out->microphys.denstab_amf[i][lc];

        isthis=-666;
        /* replace the profile in the atmosphere-data according to sza */
        iv=0;  /* No solar zenith variations for a single amf calculation */
        for (is=0; is<out->microphys.nsza_denstab; is++)
          if (fabs (out->microphys.sza_denstab[is]-out->sza_t[iv]) < 1.e-04)
	    isthis = is;

        if (isthis < 0) {
          /* sza not found in the denstab-file */
          if (!quiet) {
            fprintf (stderr, "Warning: requested sza is not specified in the denstab-file\n");
            fprintf (stderr, "  ***    Gas profile from the atmosphere file is used. \n");
          }

          /* interpolate data of the background-atmosphere */
          /* from atm-file grid onto the z_all-grid  */
          status += interpolate_density (z_atm, &(out->microphys.dens[dens_id]), nlev_atm, 
                                         z_all, nlev_all, interpol_method_gas[dens_id],
                                         dens_air_atm, dens_air_all, quiet);
        }
        else {
          /* normal case, sza found in the denstab file */
          fprintf (stderr, "sza = %6.2f\n", out->sza_t[0]);

          free(out->microphys.dens[dens_id]);
          out->microphys.dens[dens_id] = (float *) calloc (nlev_all, sizeof (float));

          for (lc=0; lc<nlev_all; lc++){
            out->microphys.dens[dens_id][lc] = out->microphys.denstab[dens_id][isthis][lc];
          }
        }

        first_dens = 0;

      } /* if (max_columns==2) */

      free(zd);
      free(dens);
      free(z_comb);
      free(dens_comb);
  
      ASCII_free_float(atm_data, rows);

    }       /* if (strlen(filename) > 0) */
    else {  /* (there is no densfile, but z-grids are different)    */

      /* interpolation from atm-file grid onto the z_all-grid  */
      status += interpolate_density (z_atm, &(out->microphys.dens[dens_id]), nlev_atm, 
                                     z_all, nlev_all, interpol_method_gas[dens_id],
                                     dens_air_atm, dens_air_all, quiet);

    } /* if (strlen(filename) > 0) */

  } /* else { ( something to do ) */ 
  
  return status;
}


/**************************************/
/* Combine two profiles               */
/**************************************/

static int combine_profiles (float **dens, float *z, int nlev, int nsza, 
                             float *dens_atm, float *z_atm, int nlev_atm, 
                             int interpol_method_dens,
                             float ***dens_comb, float **z_comb, int *nlev_comb)

{  /* _atm means z-grid of the atmosphere files */
   /* comb means combined                */
   /* output are dens_comb and z_comb    */

  int lc=0,i=0;
  int n_up=-2, n_dn=-2;

  n_up = 0;
  n_dn = nlev_atm-1;

  if ( z[0] > z_atm[0] ||  z[nlev-1] < z_atm[nlev_atm-1] ){
    fprintf (stderr, "Error in function combine profiles!\n");  
    fprintf (stderr, "z-range of dens_file: z[0]= %f, z[%d]=%f\n", z[0],nlev,z[nlev-1]);
    fprintf (stderr, "is larger than those of the background atmosphere\n"); 
    fprintf (stderr, "z-range of the atm-file: z_atm[0] = %f, z[%d] = %f\n", z_atm[0],nlev_atm-1,z_atm[nlev_atm-1] );   
    return -1;
  }

  if ( z[0] == z_atm[0] &&  z[nlev] == z_atm[nlev_atm] ){
    /* upper and lower boundary are identical, no combination nessesary */
    *nlev_comb = nlev;

    *z_comb    = calloc (*nlev_comb, sizeof (float  ) );
    *dens_comb = calloc ( nsza,      sizeof (float *) );
    for (i=0; i<nsza; i++)
      (*dens_comb)[i] = calloc (*nlev_comb, sizeof (float));

    for (lc=0; lc<*nlev_comb; lc++)
      (*z_comb)[lc]    = z[lc];

    for (i=0; i<nsza; i++) 
      for (lc=0; lc<*nlev_comb; lc++)
        (*dens_comb)[i][lc] = dens[i][lc];
 
  }
  else {

    /* get z_atm index above and below the dens-data */
    for  (lc=0; lc<nlev_atm; lc++) {
      if( z_atm[lc] <= z[0]) {        
        n_up = lc-1;     /* last index of z_atm above the z1-grid-range */
        break;
      }
    }

    for  (lc=(nlev_atm-1); lc>=0; lc--) {
      if( z_atm[lc] >= z[nlev-1]){
        n_dn = lc;       /* last index of z_atm within the z1-grid-range */
	break;
      }
    }
  
    if ( n_up < -1 ){
     fprintf (stderr, "Error in function combine_profiles!\n");  
     fprintf (stderr, "upper index not found, n_up = %d \n", n_up);   
     return -1;
    }
    if ( n_dn < 0 ){
     fprintf (stderr, "Error in function combine_profiles!\n");  
     fprintf (stderr, "lower index not found, n_down = %d \n", n_dn);   
     return -1;
    }

    (*nlev_comb) = (n_up + 1) + nlev + ((nlev_atm - 1) - n_dn);
    
    /* combine the two profiles, dens (on z) has priority */
    *z_comb    = calloc (*nlev_comb, sizeof (float  ) );
    *dens_comb = calloc ( nsza,      sizeof (float *) );
    for (i=0; i<nsza; i++)
      (*dens_comb)[i] = calloc (*nlev_comb, sizeof (float));

    /* combine z-grid */
    for (lc=0; lc<=n_up; lc++)
         (*z_comb)[lc]          =    z_atm[lc];
    for (lc=0; lc<nlev; lc++)
         (*z_comb)[lc+(n_up+1)] =    z[lc];
    for (lc=1; lc<(nlev_atm-n_dn);lc++)
         (*z_comb)[lc+(n_up+nlev)] =    z_atm[lc+n_dn];

    /* combine data */
    for (i=0; i<nsza; i++){
      for (lc=0; lc<=n_up; lc++)
        (*dens_comb)[i][lc] = dens_atm[lc];
      for (lc=0; lc<nlev; lc++)
        (*dens_comb)[i][lc+(n_up+1)] = dens[i][lc];
      for (lc=1; lc<(nlev_atm-n_dn);lc++)
        (*dens_comb)[i][lc+(n_up+nlev)] = dens_atm[lc+n_dn];
    }
  }
  
/*   for (lc=0; lc<*nlev_comb;lc++) */
/*     fprintf (stderr,"z_comb[%d] = %f, dens_comb[%i] = %11.7e\n", lc, (*z_comb)[lc], lc, (*dens_comb)[0][lc]); */

  return 0;
}

/**************************************************************/
/* Convert relative humidity [%] to concentration [1/cm3]     */
/**************************************************************/

static int read_and_convert_rh (char *filename, float **rh,
                                float *z_all, int nlev_all,
				int interpol_method_rh,
			        float *t, float *h2o,
                                int quiet)
{
  int status=0, nlev=0, lc=0;

  float *z=NULL;
  double conc=0;
  float *rh_all=NULL;
  float **rh_comb=NULL;
  float *z_comb=NULL;
  int nlev_comb=-666;

  /* read relative humidity from file */
  if (strlen(filename)>0) {

    if (!quiet)
      fprintf (stderr, " ... reading relative humidity from %s\n", filename);

    status = read_2c_file_float (filename, &z, rh, &nlev);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    if (z[0] == z_all[0] && z[nlev-1] == z_all[nlev_all-1]){
      /* interpolate rh profile to common atmospheric grid, linear */
      status += interpolate_profile (z, rh, nlev,
                                     z_all, nlev_all,
                                     interpol_method_rh, quiet);

      if (status!=0) {
        fprintf (stderr, "Error %d interpolating relative humidity to common atmospheric grid\n", status);
        return status;
      }
    }
    else {
      /* combine rh_file and atmosphere-file data*/

      /* calculate rh-file of the atmosphere-file h2o-data */
      rh_all = (float *) calloc (nlev_all, sizeof(float));
      if (rh_all == NULL) {
        fprintf (stderr, "Error allocating memory for rh_all in function read_and_convert_rh \n");
        return -1;
      }
      for (lc = 0; lc < nlev_all; lc++){
        conc = vapor_pressure (t[lc]);
        rh_all[lc] = 100.0 * h2o[lc] / conc;
      }

      /* combine rh-file data and atmosphere-file rh-data  */
      combine_profiles ( rh, z, nlev, 1,
                         rh_all, z_all, nlev_all,
                         interpol_method_rh,
                         &(rh_comb), &(z_comb), &(nlev_comb));

      free(*rh);
      (*rh) = (float *) calloc (nlev_all, sizeof(float));

      status += arb_wvn (nlev_comb, z_comb, rh_comb[0],
                         nlev_all,  z_all, *rh,
                         interpol_method_rh, 1);
      free(z_comb);
      free(rh_comb[0]);
      free(rh_comb);
      free(rh_all);
    }

  free(z);

  }
  else {  /* create dummy arrays */

    /* allocate memory */
    *rh = (float *) calloc (nlev_all, sizeof(float));
    if (*rh == NULL) {
      fprintf (stderr, "Error allocating memory for rh in function read_and_convert_rh \n");
      return -1;
    }
    for (lc = 0; lc < nlev_all; lc++)
      (*rh)[lc] = -1.0;
  }

  for (lc = 0; lc < nlev_all; lc++) {

    conc = vapor_pressure (t[lc]);

    /* if a RH is defined for this level, replace the water vapor concentration */
    if ((*rh)[lc]>=0) {
      h2o[lc] = conc * (*rh)[lc]/100.0;
    }
    /* if no RH is defined for this level, use the original water vapor concentration */
    else {
      (*rh)[lc] = 100.0 * h2o[lc] / conc;
    }
  }

  return 0;   /* o.k. */
}



/**************************************************************/
/* Read refraction index file                                 */
/* Combine with background atmosphere and                     */
/* Interpolate on the common grid: z_all                      */
/**************************************************************/

static int calculate_or_read_refind  (char *filename,
				      float **refind, float *z_all, int nlev_all,
				      int interpol_method_refind, 
                                      float *t, float *p,
                                      int quiet)

{
  int status=0, nlev=0, lc=0;

  float *z=NULL;
  float *refind_all=NULL;
  float **refind_comb=NULL;
  float *z_comb=NULL;
  int nlev_comb=-666;
  float n_0 = 1.000293; /* refraction under normal conditions */   
  float t_0 = 15.0;     /* normal temperature in celsius      */
  float p_0 = 1013.25;  /* normal pressure in mbar            */
  float ar  = 0.00366;  /* parameter for calculating refind   */


  /* read refractive index from file */
  if (strlen(filename)>0) {

    fprintf (stderr, " ... reading refractive index from %s\n", filename);

    status = read_2c_file_float (filename, &z, refind, &nlev);   
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    if (z[0] == z_all[0] && z[nlev-1] == z_all[nlev_all-1]){
      /* interpolate refind profile to common atmospheric grid */
      status += interpolate_profile (z, refind, nlev,
                                     z_all, nlev_all,
                                     interpol_method_refind, quiet);
      if (status!=0) {
        fprintf (stderr, "Error %d interpolating relative humidity to common atmospheric grid\n", status);
        return status;
      }

    }
    else {
      /* combine refind_file and atmosphere-file data*/

      /* calculate refind of the atmosphere-file data */
      refind_all = (float *) calloc (nlev_all, sizeof(float));      
      if (refind_all == NULL) {
        fprintf (stderr, "Error allocating memory for refind_all in function calculate_or_read_refind \n");
        return -1;
      }

      for (lc = 0; lc < nlev_all; lc++)
        refind_all[lc] = 1.0 + (n_0-1.0) * (1.0 + ar*t_0) / (1.0 + ar * (t[lc]-273.15)) * p[lc] / p_0; 

      /* combine refind-file data and atmosphere-file refind-data  */
      combine_profiles ( refind, z, nlev, 1,
                         refind_all, z_all, nlev_all,
                         interpol_method_refind,
                         &(refind_comb), &(z_comb), &(nlev_comb));

      free(*refind);
      (*refind) = (float *) calloc (nlev_all, sizeof(float));

      status += arb_wvn (nlev_comb, z_comb, refind_comb[0],
                         nlev_all,  z_all, *refind,
                         interpol_method_refind, 1);

      free(z_comb);
      free(refind_comb[0]);
      free(refind_comb);
      free(refind_all);
    }
    free(z);
  }
  else {  /* if (strlen(filename)>0) => no refind-input-file */

    /* allocate memory */
    /* free(*refind); */     
    *refind = (float *) calloc (nlev_all, sizeof(float));
    if (*refind == NULL) {
      fprintf (stderr, "Error allocating memory for refind in function calculate_or_read_refind \n");
      return -1;
    }
    for (lc = 0; lc < nlev_all; lc++)
      (*refind)[lc] = 1.0 + (n_0-1.0) * (1.0 + ar*t_0) / (1.0 + ar * (t[lc]-273.15)) * p[lc] / p_0;
  }

  return 0;   /* o.k. */
}



/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z]. Memory for dtau[] is allocated automatically.              */ 
/***********************************************************************************/

static int read_molecular_absorption (char* filename, 
				      float **zd, int *nlyr,
				      int quiet,
				      double **dtau_mono, 
				      float ***dtau_spec, float **lambda, int *nlambda,
				      int *monochromatic) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int lu=0, iv=0;

  int fulfilled=0;

  float **data=NULL;

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_nlyr=0, idd_nwvl=0;
  int id_wvl=0, id_z=0, id_tau=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t taustart[2] = {0,0};
  size_t taucount[2] = {0,0};

  double *tmp_zd=NULL, *tmp_lambda=NULL;

  size_t dimlen=0;

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    if (!quiet) {
      fprintf (stderr, "*** %s is not a netcdf file, trying to open as ASCII instead.\n", filename);
      fprintf (stderr, "*** It is recommended to use the netCDF format because access is much faster.\n");
    }
  }
  else {
    if (!quiet)
      fprintf (stderr, " ... reading profiles from netcdf file %s\n", filename);

    /* get dimension id for "nlyr" */
    status = nc_inq_dimid (ncid, "nlyr", &idd_nlyr);
    
    /* get dimension length for "nlyr" */
    status = nc_inq_dimlen (ncid, idd_nlyr, &dimlen);
    *nlyr = dimlen;

    /* get dimension id for "nwvl" */
    status = nc_inq_dimid (ncid, "nwvl", &idd_nwvl);
    
    /* get dimension length for "nwvl" */
    status = nc_inq_dimlen (ncid, idd_nwvl, &dimlen);
    *nlambda = dimlen;

    /* read levels */
    tmp_zd = calloc (*nlyr+1, sizeof(double));
    *zd    = calloc (*nlyr+1, sizeof(float));

    /* get variable id for "z" */
    status = nc_inq_varid (ncid, "z", &id_z);

    count[0] = *nlyr+1;

    /* read "z" */
    status = nc_get_vara_double (ncid, id_z, start, count, tmp_zd);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    
    /* read wavelengths */
    tmp_lambda = calloc (*nlambda, sizeof(double));
    *lambda    = calloc (*nlambda, sizeof(float));

    /* get variable id for "wvl" */
    status = nc_inq_varid (ncid, "wvl", &id_wvl);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }
    
    /* read "wvl" */
    count[0] = *nlambda;

    status = nc_get_vara_double (ncid, id_wvl, start, count, tmp_lambda);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }


    /* get variable id for "tau" */
    status = nc_inq_varid (ncid, "tau", &id_tau);

    /* read optical thickness */
    taustart[1] = 0;  /* start with first optical thickness */
    taucount[0] = 1;  /* read one layer at a time           */
    
    *dtau_spec = calloc (*nlyr, sizeof(float *));
    for (lu=0; lu<*nlyr; lu++) {
      (*dtau_spec)[lu] = calloc (*nlambda, sizeof(float));

      taustart[0] = lu;
      taucount[1] = *nlambda;

      status = nc_get_vara_float (ncid, id_tau, taustart, taucount, (*dtau_spec)[lu]);
      if (status!=NC_NOERR) {
	fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
	return status;
      }
    }

    /* copy data to final destination and free temporary memory */
    for (lu=0; lu<=*nlyr; lu++)
      (*zd)[lu] = tmp_zd[lu];

    for (iv=0; iv<*nlambda; iv++)
      (*lambda)[iv] = tmp_lambda[iv];

    free (tmp_zd); free (tmp_lambda);

    /* this is obviously a wavelength-dependent cross section */
    *monochromatic=0;

    return 0;
  }

#endif

  if (!quiet)
    fprintf (stderr, " ... reading profiles from ASCII file %s\n", filename);

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  /* a rough format check */
  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  if (min_columns!=max_columns) {
    fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	     filename, min_columns, max_columns);
    return -1;
  }


  /* check if first column sorted in descending order;    */
  /* if yes, we assume old format: two columns z and dtau */
  fulfilled=1;
  for (lu=0; lu<rows-1; lu++)
    if (data[lu][0] < data[lu+1][0])
      fulfilled = 0;
  
  if (fulfilled) {

    /* here we assume that the file comes in the old format,  */
    /* two columns z and dtau; in this case, the first column */
    /* must equal the atmospheric profile                     */

    *monochromatic=1;

    *nlyr = rows-1;
    *zd        = calloc (*nlyr+1, sizeof (float));
    *dtau_mono = calloc (*nlyr,   sizeof (double));

    for (lu=0; lu<*nlyr+1; lu++)
      (*zd)[lu] = data[lu][0];

    for (lu=0; lu<*nlyr; lu++)
      (*dtau_mono)[lu] = data[lu+1][1];
  }
  else {
    /* now we need to assume that we deal with a new format */
    /* where the first row must be the altitude levels      */
    
    /* check if first column sorted in ascending order; */
    /* if yes, we assume wavelength-dependent format    */
    fulfilled=1;
    for (iv=1; iv<rows-1; iv++)
      if (data[iv][0] > data[iv+1][0])
	fulfilled = 0;

    if (!fulfilled) {
      fprintf (stderr, "Error reading %s\n", filename);
      return -1;
    }
    
    /* now we know that the format is correct, and the only thing   */
    /* we need to do is to copy the data to their final destination */
    
    *monochromatic=0;

    *nlyr    = min_columns-1;
    *nlambda = rows-1;

    /* allocate memory */
    *zd = calloc (*nlyr+1, sizeof (float));
    *lambda    = calloc (*nlambda, sizeof (float));
    *dtau_spec = calloc (*nlyr,    sizeof (float *));
    for (lu=0; lu<*nlyr; lu++)
      (*dtau_spec)[lu] = calloc (*nlambda, sizeof(float));


    for (lu=0; lu<=*nlyr; lu++)
      (*zd)[lu] = data[0][lu];

    for (iv=0; iv<*nlambda; iv++) {
      (*lambda)[iv] = data[iv+1][0];
      
      for (lu=0; lu<*nlyr; lu++)
	(*dtau_spec)[lu][iv] = data[iv+1][lu+1];
    }
  }

  /* free memory */
  ASCII_free_float (data, rows);

  return 0;
}


/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z].                                                            */
/***********************************************************************************/

static int read_molecular_absorption_z (char* filename, int quiet,
					float **zd, int *nlyr) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int lu=0, iv=0;

  int fulfilled=0;

  
  float **data=NULL;
  

#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_nlyr=0;
  int id_z=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t dimlen=0;

  double *z=NULL;

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {
    /* get dimension id for "nlyr" */
    status = nc_inq_dimid (ncid, "nlyr", &idd_nlyr);
    
    /* get dimension length for "nlyr" */
    status = nc_inq_dimlen (ncid, idd_nlyr, &dimlen);
    *nlyr = dimlen;

    /* read levels */
    z   = calloc (*nlyr+1, sizeof(double));
    *zd = calloc (*nlyr+1, sizeof(float));

    /* get variable id for "z" */
    status = nc_inq_varid (ncid, "z", &id_z);

    count[0] = *nlyr+1;

    /* read "z" */
    status = nc_get_vara_double (ncid, id_z, start, count, z);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* copy profile to final destination and free temporary memory */
    for (lu=0; lu<=*nlyr; lu++)
      (*zd)[lu] = z[lu];
    free (z);

    nc_close (ncid);
    
    return 0;
  }

#endif

  if (!quiet)
    fprintf (stderr, " ... reading altitudes from ASCII file %s\n", filename);

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  /* a rough format check */
  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  if (min_columns!=max_columns) {
    fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	     filename, min_columns, max_columns);
    return -1;
  }


  /* check if first column sorted in descending order;    */
  /* if yes, we assume old format: two columns z and dtau */
  fulfilled=1;
  for (lu=0; lu<rows-1; lu++)
    if (data[lu][0] < data[lu+1][0])
      fulfilled = 0;
  
  if (fulfilled) {
    *nlyr = rows-1;
    *zd = calloc (rows, sizeof (float));

    for (lu=0; lu<rows; lu++)
      (*zd)[lu] = data[lu][0];
  }
  else {
    /* now we need to assume that we deal with a new format */
    /* where the first row must be the altitude levels      */
    
    /* check if first column sorted in ascending order; */
    /* if yes, we assume wavelength-dependent format    */
    fulfilled=1;
    for (iv=1; iv<rows-1; iv++)
      if (data[iv][0] > data[iv+1][0])
	fulfilled = 0;

    if (!fulfilled) {
      fprintf (stderr, "Error reading %s\n", filename);
      return -1;
    }
    
    *nlyr = min_columns-1;
    *zd = calloc (min_columns, sizeof (float));
    for (lu=0; lu<min_columns; lu++)
      (*zd)[lu] = data[0][lu];
  }

  /* free memory */
  ASCII_free_float (data, rows);

  return 0;
}




/***********************************************************************************/
/*  Read the molecular absorption file which can either be a two-column file       */
/*  containing altitude [km] and optical thickness of each layer or a matrix       */
/*  tau[wavelength, z]. Check if the levels equal those defined in zd[]. Memory    */
/*  for dtau[] is allocated automatically.                                         */ 
/***********************************************************************************/

int read_molecular_absorption_lambda (char* filename, int quiet,
				      float **lambda, int *nlambda, int *monochromatic) 
{
  int rows=0, min_columns=0, max_columns=0, status=0;
  int iv=0;

  int fulfilled=0;

  float **data=NULL;
  
#if HAVE_LIBNETCDF
  /* try to open the CDF file */

  char cdf_filename[FILENAME_MAX]="";
  int ncid=0;

  int idd_nwvl=0;
  int id_wvl=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  size_t dimlen=0;

  double *wvl=NULL;

  strcpy (cdf_filename, filename);
  strcat (cdf_filename, ".cdf");

  /* open netcdf file */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status==NC_NOERR) {
    /* get dimension id for "nwvl" */
    status = nc_inq_dimid (ncid, "nwvl", &idd_nwvl);
    
    /* get dimension length for "nwvl" */
    status = nc_inq_dimlen (ncid, idd_nwvl, &dimlen);
    *nlambda = dimlen;

    /* read levels */
    wvl     = calloc (*nlambda, sizeof(double));
    *lambda = calloc (*nlambda, sizeof(float));

    /* get variable id for "wvl" */
    status = nc_inq_varid (ncid, "wvl", &id_wvl);

    count[0] = *nlambda;

    /* read "wvl" */
    status = nc_get_vara_double (ncid, id_wvl, start, count, wvl);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, cdf_filename);
      return status;
    }

    /* copy profile to final destination and free temporary memory */
    for (iv=0; iv<*nlambda; iv++)
      (*lambda)[iv] = wvl[iv];
    free (wvl);

    *monochromatic=0;

    nc_close (ncid);

    return 0;
  }

#endif

  if (!quiet)
    fprintf (stderr, " ... reading wavelengths from ASCII file %s\n", filename);

  /* read altitude and optical depth from file */
  status = ASCII_file2float (filename,   
			     &rows,        
			     &max_columns, 
			     &min_columns, 
			     &data);

  if (status!=0) {
    fprintf (stderr, "Error %d reading %s\n", status, filename);
    return status;
  }


  /* a rough format check */
  if (min_columns<2) {
    fprintf (stderr, "Error, found less than 2 columns in %s\n", filename);
    return -1;
  }

  if (min_columns!=max_columns) {
    fprintf (stderr, "Error, inconsistent number of columns in %s, %d - %d\n",
	     filename, min_columns, max_columns);
    return -1;
  }


  /* check if first column sorted in descending order;    */
  /* if yes, we assume old format: two columns z and dtau */
  fulfilled=1;
  for (iv=0; iv<rows-1; iv++)
    if (data[iv][0] < data[iv+1][0])
      fulfilled = 0;
  
  if (fulfilled)
    *monochromatic=1;
  else {
    /* now we need to assume that we deal with a new format */
    /* where the first column must be the wavelengths       */
    
    *monochromatic=0;

    fprintf (stderr, " ... now testing as wavelength-dependent absorption file\n");
    fflush (stderr);
    
    /* check if first column sorted in ascending order; */
    /* if yes, we assume wavelength-dependent format    */
    fulfilled=1;
    for (iv=1; iv<rows-1; iv++)
      if (data[iv][0] > data[iv+1][0])
	fulfilled = 0;

    if (!fulfilled) {
      fprintf (stderr, "Error reading %s\n", filename);
      return -1;
    }
    
    *nlambda = rows-1;
    *lambda = calloc (rows-1, sizeof (float));
    for (iv=0; iv<*nlambda; iv++)
      (*lambda)[iv] = data[iv+1][0];
  }

  /* free memory */
  ASCII_free_float (data, rows);

  return 0;
}



/**************************************************************/
/* Calculate atmospheric column in DU.                        */
/**************************************************************/

static int column (float *dens, float *zd, int nlev, float altitude, 
                   float *column, int interpol_method, float *dens_air)
{

  float deltaz=0;
  int lc=0, ialt=0;
  float density=0.0;
  
  for (ialt=0; ialt<nlev; ialt++)
    /*    if (altitude == zd[ialt])  which one is correct ??? */
    if (altitude >= zd[ialt])
      break;
    
  if (ialt == nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid;\n", altitude);
    fprintf (stderr, "exiting column()!\n");
    return -1;
  }

  switch (interpol_method) {

  case INTERP_METHOD_LINMIX:
  case INTERP_METHOD_SPLINE:
  case INTERP_METHOD_LOG_SPLINE:
    for (lc=0; lc<=ialt; ++lc)
      dens[lc] = dens[lc]/dens_air[lc]; /* conversion to mixing ratio !!! */
    break;
  case INTERP_METHOD_LINEAR:
  case INTERP_METHOD_LOG:
    break;
  default:
    fprintf (stderr, "Error, unknown interpolation method input.");
    return -1;  
  }

  *column=0;

  for (lc=1; lc<=ialt; ++lc) {
    deltaz = (zd[lc-1] - zd[lc]) * 1.0E5;

    switch (interpol_method) {
    case INTERP_METHOD_LINEAR:
      density = 0.5 * (dens[lc-1] + dens[lc]);   
      break;
    case INTERP_METHOD_LOG:
      density = log_average(dens[lc-1],dens[lc]);
      break;
    case INTERP_METHOD_LINMIX:
    case INTERP_METHOD_SPLINE:        /* no     spline interpolation implemented jet,  */   
    case INTERP_METHOD_LOG_SPLINE:    /* no log_spline interpolation implemented jet,  */ 
                                      /* use instead linmix assumption                 */
      density = linmix_average(dens[lc-1],dens[lc],dens_air[lc-1],dens_air[lc]);
                                      /* dens[lc - 1] and dens[lc] are mixing ratios here !!! */
      break;
    default:
      fprintf (stderr, "Error, unknown interpolation method input.");
      return -1;  
    }
    *column += density * deltaz;
  }



  switch (interpol_method) {

  case INTERP_METHOD_LINMIX:
  case INTERP_METHOD_SPLINE:
  case INTERP_METHOD_LOG_SPLINE:
    for (lc=0; lc<=ialt; ++lc) 
      dens[lc] = dens[lc]*dens_air[lc]; /* conversion back to number density */
    break;
  case INTERP_METHOD_LINEAR:
  case INTERP_METHOD_LOG:
    break;
  default:
    fprintf (stderr, "Error, unknown interpolation method input.");
    return -1;  
  }

  *column /= 2.687E16;   /* number density column [cm-2] to DU */
  
  return 0;
}


/**************************************************************/
/* Scale a gas column to a given value.                       */
/**************************************************************/

static int scale_profile (float *dens, float *zd, int nlev, float newcol, float altitude, float *factor, 
                          int interpol_method, float *dens_air)
{
  int lc=0, status=0;
  float oldcol=0;

  /* determine actual ozone column */
  status = column (dens, zd, nlev, altitude, &oldcol, interpol_method, dens_air);
  if (status!=0) {
    fprintf (stderr, "Error %d calculating old column in scale_profile(), exiting\n", status);
    return status;
  }

  /* determine scale factor */
  if ( oldcol > 0.0 )
    *factor = newcol / oldcol;
  else
    *factor = 0;

  /* scale profile */
  for (lc = 0; lc<nlev; ++lc)
    dens[lc] *= (*factor);
  
  return 0;
}



/**************************************************************/
/* Calculate preciptable water [kg / m2].                     */
/**************************************************************/

static int precip_water (float *dens, float *zd, int nlev, float altitude, float *precip, int interpol_method, float *dens_air)
{

  float deltaz=0;
  int lc=0, ialt=0;
  float density=0.0;
  
  for (ialt=0; ialt<nlev; ialt++)
    if (altitude == zd[ialt])
      break;
    
  if (ialt == nlev) {
    fprintf (stderr, "Error, user-defined altitude %f not found in altitude grid;\n", altitude);
    fprintf (stderr, "exiting column()!\n");
    return -1;
  }


  if (( interpol_method == INTERP_METHOD_LINMIX)   || 
      ( interpol_method == INTERP_METHOD_SPLINE)   ||
      ( interpol_method == INTERP_METHOD_LOG_SPLINE )) {
    for (lc=0; lc<=ialt; ++lc){ 
      dens[lc] = dens[lc]/dens_air[lc]; /* conversion to mixing ratio !!! */
    }
  } 

  *precip=0;

  for (lc=1; lc<=ialt; ++lc) {
    deltaz = (zd[lc-1] - zd[lc]) * 1.0E3; /* conversion from km to m => 1.0E3 */

    switch (interpol_method) {
    case INTERP_METHOD_LINEAR:
      density = 0.5 * (dens[lc-1] + dens[lc]);
      break;
    case INTERP_METHOD_LOG:
      density = log_average(dens[lc-1],dens[lc]);
      break;
    case INTERP_METHOD_LINMIX:
    case INTERP_METHOD_SPLINE:        /* no     spline interpolation implemented jet,  */   
    case INTERP_METHOD_LOG_SPLINE:    /* no log_spline interpolation implemented jet,  */ 
                                      /* use instead linmix assumption                 */
      density = linmix_average(dens[lc-1],dens[lc],dens_air[lc-1],dens_air[lc]);
                                      /* dens[lc - 1] and dens[lc] are mixing ratios here !!! */    
      break;
    default:
      fprintf (stderr, "Error, unknown interpolation method input.");
      return -1;  
    }

    *precip += density * deltaz ;
  }
  
  if (( interpol_method == INTERP_METHOD_LINMIX)   || 
      ( interpol_method == INTERP_METHOD_SPLINE)   ||
      ( interpol_method == INTERP_METHOD_LOG_SPLINE )) {
    for (lc=0; lc<=ialt; ++lc) 
      dens[lc] = dens[lc]*dens_air[lc]; // conversion back to number density !!!
  }

  *precip *= (M_H2O / N_MOL_CKDFU * 1e3);
  
  return 0;
}


/**************************************************************/
/* Scale precipitable water to a given value.                 */
/**************************************************************/

static int scale_precip (float *dens, float *zd, int nlev, float precip, float altitude, float *factor, int interpol_method, float *dens_air)
{
  int lc=0, status=0;
  float precipwat=0;

  /* determine actual precipitable water */
  status = precip_water (dens, zd, nlev, altitude, &precipwat, interpol_method, dens_air);
  if (status!=0) {
    fprintf (stderr, "Error %d returned by precip_water(), exiting scale_precip()\n", status);
    return status;
  }

  /* determine scale factor */
  *factor = precip / precipwat;

  /* scale profile */
  for (lc = 0; lc<nlev; ++lc)
    dens[lc] *= (*factor);
  
  return 0;
}



/**************************************************************/
/* Convert relative humidity [%] to concentration [1/cm3]     */
/**************************************************************/

static double vapor_pressure (double t)
{
  /*
  double c1=-6096.9385;
  double c2=16.635794;
  double c3=-2.711193e-2;
  double c4=1.673952e-5;
  double c5=2.433502;
  double boltzmann = 1.38065e-23;
  */

  double avogadro  = 6.022142e23;
  
  double tzero = 273.15;
  double a = tzero/t;
  double mh2o = M_H2O/avogadro;


  /* Sonntag, Meteorol. Zeitschr, n.f. 3, 51-66, 1994                                 */
  /* return 100.0 / boltzmann / t * exp (c1/t +c2 +c3*t +c4*t*t + c5*log(t)) / 1.0e6; */
  
  /* SBDART, -> CRC handbook */
  return a * exp (18.916758 - a * (14.845878 + a*2.4918766)) / mh2o / 1.0e6;
}



/*******************************************/
/* read atmospheric grid from generic ck   */
/* netcdf file                             */
/*******************************************/

static int read_nc_atmosphere_grid (float **zd, int *nlev, char *filename, int quiet)
{
#if HAVE_LIBNETCDF
  int status=0;

  size_t dimlen=0;

  size_t start[1] = {0};
  size_t count[1] = {0};

  int ncid=0;
  int idd_nlev=0, id_z=0;

  /* determine number of atmospheric levels */
  status = nc_open (filename, NC_NOWRITE, &ncid);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
    return status;
  }
  
  /* get dimension id for "nlev" */
  status = nc_inq_dimid (ncid, "nlev", &idd_nlev);
  
  /* get dimension length for "nlev" */
  status = nc_inq_dimlen (ncid, idd_nlev, &dimlen);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading \"nlev\" from %s\n", status, filename);
    return status;
  }

  *nlev = dimlen;
  *zd   = calloc (*nlev, sizeof (float));

  /* get variable id for "z" */
  status = nc_inq_varid (ncid, "z", &id_z);
  
  /* read whole field */
  start[0] = 0;
  count[0] = *nlev;

  /* read altitude "z" */
  status = nc_get_vara_float (ncid, id_z, start, count, *zd);
  if (status!=NC_NOERR) {
    fprintf (stderr, "Error %d reading z from %s\n", status, filename);
    return status;
  }

  nc_close (ncid);
  return 0;

#else
  fprintf (stderr, "Error, netcdf required by read_nc_atmosphere_grid();\n");
  fprintf (stderr, "please install netcdf and recompile libRadtran!\n");
  return -1;
#endif
}

